// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_bcast_alphatree.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of bcast using binomial algorithm
// $Id: ics_bcast_alphatree.cpp,v 1.5 2003/07/18 11:26:00 olobry Exp $
//
// ==========================================================================
#include <math.h>
#include <ics_conf.h>
#include <ics_bcast_alphatree.h>


namespace ICS {

/* Diffusion du message (buffer, count), en utilisant l'identificateur tag,
    a tous les processus de rank contenu dans ranks[0] ... ranks[sz-1]. 
    Le processus qui appelle cette fonction a le rank ranks[index_self]. 
    Le processus qui possede la donnee a le rank ranks[index_src]. 
    Tous les processus qui participe a cette diffusion doivent appeler 
    cette fonction. buffer est alloue sur tous les processus, initialement 
    seul le processus de rank ranks[index_src] possede une donnee initialisee. 
*/


// -----------------------------------------------------------------------------
BcastAlphaTree::BcastAlphaTree( double alpha )
  : _alpha(alpha)
{ }


// -----------------------------------------------------------------------------
/** */
struct SAlphaTree_continuation {
  BcastAlphaTree* serv;
  void**                     buffer;
  size_t*                    count;
  const ACOM::Format*        fmt;
  int                        sz_ranks;
  int*                       ranks;
  int                        index_self;
  int                        index_src;
  int                        tag;
  void                       (*cb_func)(void*, void*);
  void*                      cb_arg1;
  void*                      cb_arg2;
};


/** */
void BcastAlphaTree::continuation( void* a1, void* a2 )
{
  /* here message is arrived: either resend before call notification,
     or call notification before.
     We assume that data could changed in cb_func, so we postponed notification
     at the end.
  */

  /* compute the parition of the set of node */
  SAlphaTree_continuation* cc = (SAlphaTree_continuation*)a1;
  int monRang = cc->index_self;
  int n = cc->sz_ranks;
  int parent = 0;
  int nOthers;
  int threshold;

  // cached copy of alpha
  double alpha = cc->serv->_alpha;

//std::cout << _net->rank() << "::my rank=" << ranks[index_self] << std::endl;

  while(monRang != 0) {
    nOthers = (int)round(alpha*n);
    threshold = n-nOthers;
    if(monRang == threshold) {
      n = nOthers;
      monRang -= threshold;
    } else {
      if(monRang < threshold)
        n = threshold;
      else {
        n = nOthers;
        parent += threshold;
        monRang -= threshold;
      }
    }
  }

  /* count the number of send to do */
  int nn = n;
  int nbsend = 0;
  while(nn > 1) {
    nn -= (int)round(alpha*nn);
    ++nbsend;
  }
  if (nbsend ==0) goto notify;
  
  {  /* Form the message */
    ACOM::WriteAndSignal<BcastAlphaTree> ws (cc->serv->_net, cc->serv, 0, nbsend);
    cc->serv->postsend_data( ws.get_request(), cc->tag, cc->fmt, *cc->buffer, *cc->count );
  
    while(n > 1) {
      n -= (int)round(alpha*n);
      /* send messag to ranks[ (index_self + n) % sz] */
      int dest = cc->ranks[ (cc->index_self + n) % cc->sz_ranks];
      ACOM::Network::OutChannel* channel = cc->serv->_net->get_default_route( dest );
      /* Send it */
      ws.post(channel);
    }
  }  

notify:
  if (cc->cb_func !=0)
    (*cc->cb_func)( cc->cb_arg1, cc->cb_arg2 );
  delete cc;
}
 


// -----------------------------------------------------------------------------
/** */
BcastAlphaTree::handler BcastAlphaTree::post_bcast(
    void*&                     buffer, 
    size_t&                    count,
    const ACOM::Format*        fmt,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_src, 
    int                        tag
  )
{
//std::cout << "IN " << __PRETTY_FUNCTION__ << std::endl;
  ++_nbcalled;
  if (sz_ranks <=1) return 0;
  
  SAlphaTree_continuation* cc = new SAlphaTree_continuation;
  cc->serv       = this;
  cc->sz_ranks   = sz_ranks;
  cc->ranks      = ranks;
  cc->index_self = index_self;
  cc->index_src  = index_src;
  cc->tag        = tag;
  cc->fmt        = fmt;
  cc->buffer     = &buffer;
  cc->count      = &count;
  cc->cb_func    = 0;
  cc->cb_arg1    = 0;
  cc->cb_arg2    = 0;

  if(index_self != index_src) 
  {
    /* should receive a message */
    PTH::CriticalGuard cg(_access); // - begin critical section    
    postrecv_data( tag, true, fmt, &buffer, &count, 0, 0, 0 );
    /* continuation will do the next step */
    return (handler)cc;
  }

  /* else, I'm master do it now: */  
  BcastAlphaTree::continuation( cc, 0);
  return 0;
}



// -----------------------------------------------------------------------------
/** */
void BcastAlphaTree::wait_bcast(
  handler wc
)
{
  if (wc ==0) return; // master
  SAlphaTree_continuation* cc = (SAlphaTree_continuation*)wc;
  { PTH::CriticalGuard cg(_access); // - begin critical section    
    com_context* ctxt = wait_data( cc->tag);
    deallocate_cc( ctxt );
  }
  /* I'm slave, resend data if any */  
  BcastAlphaTree::continuation( cc, 0);  
}



// -----------------------------------------------------------------------------
/** */
void BcastAlphaTree::async_bcast(
    void*&                     buffer, 
    size_t&                    count,
    const ACOM::Format*        fmt,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_src, 
    int                        tag,
    void                       (*cb_func)(void*, void*),
    void*                      cb_arg1,
    void*                      cb_arg2
  )
{
  ++_nbcalled;
  if (sz_ranks <=1) return;
    
  SAlphaTree_continuation* cc = new SAlphaTree_continuation;
  cc->serv       = this;
  cc->sz_ranks   = sz_ranks;
  cc->ranks      = ranks;
  cc->index_self = index_self;
  cc->index_src  = index_src;
  cc->tag        = tag;
  cc->fmt        = fmt;
  cc->buffer     = &buffer;
  cc->count      = &count;
  cc->cb_func    = cb_func;
  cc->cb_arg1    = cb_arg1;
  cc->cb_arg2    = cb_arg2;

  if(index_self != index_src) 
  {
    /* should receive a message */
    PTH::CriticalGuard cg(_access); // - begin critical section    
    postrecv_data( tag, false, fmt, &buffer, &count, &BcastAlphaTree::continuation, cc, 0 );
    /* continuation will do the next step */
    return;
  }

  /* else do it now: */  
  BcastAlphaTree::continuation( cc, 0);
}



}
