// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_bcast_pipeline1.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of bcast using pipeline algorithm
// $Id: ics_bcast_pipeline1.cpp,v 1.4 2003/07/15 17:22:44 thierry Exp $
//
// ==========================================================================
#include <ics_conf.h>
#include <ics_bcast_pipeline1.h>
#include <iostream>
#ifdef PTH_USE_PTHREAD_DARWIN
#include <stdlib.h>
#else
#include <alloca.h>
#endif

/* Note: The algorithm is the following.
 * The message to broadcast (buffer,count,fmt) is developped into an iovector.
 * The source node send the description of header part of the message :
 * the network header + immediate data. Then all other nodes have the 
 * format and could reserved memory for defered data.
 * Defered data is an iovector sent by blocs of iovector such that each bloc
 * has a size less than a fixed size.
 * The ccom_context data structure is used as the following:
 - count_recv in header packet is the count of sended data
 - data_post are pointers to IOVectEntry
 - count_recv is the number of IOVectEntry
 */
namespace ICS {


// -----------------------------------------------------------------------------
/* */
void BcastPipeline1::initialize( ACOM::Network* net )
{ 
  BcastService::initialize(net);
  _iovA.init( sizeof(ACOM::IOVectCell)+(_nb_entries*sizeof(ACOM::IOVectEntry)),
              _iovect_threshold );
}


// -----------------------------------------------------------------------------
/* The algorithm use the fact that ::call re-send the incomming bloc
*/
BcastPipeline1::handler BcastPipeline1::post_bcast(
    void*&                     buffer, 
    size_t&                    count,
    const ACOM::Format*        fmt,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_src, 
    int                        tag
  )
{
  ++_nbcalled;
  if (sz_ranks <= 1) return 0;

  /* node where to send */
  int indextosend = (index_self+1)%sz_ranks;

  /* no broadcast : one node ! */
  if (indextosend == index_self)
    return 0;

  /* get a channel */
  ACOM::Network::OutChannel* channel = _net->get_default_route( ranks[indextosend] );

  /* format id */
  int fid = fmt->get_id();

  if (index_self == index_src) 
  { 
#if 0
    std::cout << "I am " << ranks[index_self] << " and I initiate communication" << std::endl;
#endif

    /* I'm source : begin to send the message using packets of size size_bloc.
       Each packet has a serial number for this message with tag 'tag'.
    */
    ACOM::Network::Request* req = _net->allocate_request();
    if (count >0) {
      ACOM::OStream msg ( req->get_datastream(), req->get_iovstream() );
      fmt->write(msg, buffer, count);
    }
      
    /* close the sequences */
    req->get_datastream()->packEos();
    req->get_iovstream()->putEos();
    
    size_t size_h  = req->get_datastream()->getFillSize();
    size_t iovsz_h = req->get_datastream()->getBlockSize();
    size_t size_d  = req->get_iovstream()->getByteSize();
    size_t iovsz_d = req->get_iovstream()->getBlockSize();
    size_t iovsz = iovsz_h+iovsz_d;

    /* Store in iovdlen the length of defered iovector entries. 
     */
    size_t* iovdlen = (size_t*)::alloca( sizeof(size_t) * (4 + iovsz) );
    ACOM::IOVectEntry* iov_send = (ACOM::IOVectEntry*)::alloca( sizeof(ACOM::IOVectEntry) * (iovsz) );
    iovdlen[0] = size_h;
    iovdlen[1] = iovsz_h;
    iovdlen[2] = size_d;
    iovdlen[3] = iovsz_d;

    int i=4;
    ACOM::IOVectEntry* curr = iov_send;
    AC::IOVectSeq::IteratorEntry beg_ie = req->get_datastream()->beginEntry();
    AC::IOVectSeq::IteratorEntry end_ie = req->get_datastream()->endEntry();
    for( ; beg_ie != end_ie; ++beg_ie, ++i, ++curr)
    {
      iovdlen[i] = beg_ie->len;
      *curr = *beg_ie;
    }

    beg_ie = req->get_iovstream()->beginEntry();
    end_ie = req->get_iovstream()->endEntry();
    for( ; beg_ie != end_ie; ++beg_ie, ++i, ++curr)
    {
      iovdlen[i] = beg_ie->len;
      *curr = *beg_ie;
    }
#if 0
std::cout << "Send iovdlen, sz=" << 4+iovsz << std::endl;
for (int k=0; k<iovsz+4; ++k)
  std::cout << iovdlen[k] << ' ';
std::cout << std::endl;      
#endif

    int serial = 0;
    /* Form the header (serial ==0) for all packets  */
    ACOM::WriteAndSignal<BcastPipeline1> ws (_net, this, 0);
    ws->pack(&serial, sizeof(int));
    ws->pack(&tag, sizeof(int));
    ws->pack(&fid, sizeof(int));
    ws->pack(&count, sizeof(size_t));
    ws->put(iovdlen,  sizeof(size_t)*(4 + iovsz));
    ws.post(channel);
    ++serial;
    
    /* Send iovector splitted by bloc for the immediate data */
    send_iov( channel, serial, tag, iov_send, iovsz_h );
  
    /* Send iovector splitted by bloc for the defered data */
    send_iov( channel, serial, tag, iov_send+iovsz_h, iovsz_d );
  
    /* free the request */
    req->finish();
    _net->deallocate_request( req );
  } 
  else 
  {
    /* should receive a message */
    PTH::CriticalGuard cg(_access); // - begin critical section

    /* I'm a receiver : wait the first packet that contains information about
      the data to received.
    */
    int serial        = 0;
    int serial_resent = 0;
    com_context* ctxt = 0;

    postrecv_iov( serial, tag, 0, (size_t)-1 );
    ctxt = wait_iov( 0, tag );
    if (ctxt->count_recv != count) {
      if (buffer !=0) {
        fmt->destroy( buffer, count);
        count = ctxt->count_recv;
        buffer = 0;
      }
    }
    if (buffer ==0) {
      buffer = fmt->create( count );
    }
    
    /* keep alias for some buffer */
    ACOM::IOVectEntry* iov_recv = (ACOM::IOVectEntry*)buffer;
    size_t* iovdlen = (size_t*)iov_recv[0].data;
    size_t size_h   = iovdlen[0];
    size_t iovsz_h  = iovdlen[1];
    size_t size_d   = iovdlen[2];
    size_t iovsz_d  = iovdlen[3];
    size_t iovsz    = iovsz_h + iovsz_d;
#if 0
std::cout << "Recv iovdlen, sz=" << 4+iovsz << std::endl;
for (int k=0; k<iovsz+4; ++k)
  std::cout << iovdlen[k] << ' ';
std::cout << std::endl;      
#endif

    if (iov_recv[0].len != (iovsz + 4) * sizeof(size_t)) {
      std::cout << "Error bad received size: iovlength for defered data" << std::endl;
    }

    /* deallocate header ctxt + curr */
    deallocate_cc( ctxt );

    /* If I am not last node, send the bloc to the next node  */
    if (indextosend != index_src)
    {
      _access.unlock();
      /* Form the message */
      ACOM::WriteAndSignal<BcastPipeline1> ws (_net, this, 0);
      ws->pack(&serial_resent, sizeof(int));
      ws->pack(&tag, sizeof(int));
      ws->pack(&fid, sizeof(int));
      ws->pack(&count, sizeof(size_t));
      ws->put(iovdlen,  sizeof(size_t)*(4 + iovsz));
      ws.post(channel);
      ++serial_resent;
#if 0
std::cout << "Resent header to node=" << ranks[indextosend] << std::endl;
#endif
      _access.lock();
    }
    
    /* sequence for all immediate data packet */
    ACOM::IOVectSeq  im_data_seq;
    im_data_seq.init(1, &_iovA);

    /* iovdlen contains the description of the message : size_h |size_d |iovsz_d | iovsz[i] */
    iov_recv = (ACOM::IOVectEntry*)::alloca( sizeof(ACOM::IOVectEntry) * (iovsz_h + iovsz_d) );
    for (size_t i =0; i < iovsz_h; ++i)
    {
      iov_recv[i].len  = iovdlen[4+i];
      iov_recv[i].data = _net->allocate_data( iov_recv[i].len );
      im_data_seq.put( &iov_recv[i] );
    }
    for (size_t i = size_h; i < iovsz_h + iovsz_d; ++i)
    {
      iov_recv[i].len  = iovdlen[4+i];
      iov_recv[i].data = 0;
    }

    /* post recv iovector splitted by bloc for the immediate data */
    postrecv_iov( serial, tag, iov_recv, iovsz_h );
  
    /* wait all message from 1 to serial */
    for (int i=1; i< serial; ++i)
    {
      ctxt = wait_iov( i, tag );
      ACOM::IOVectEntry* iov_data = (ACOM::IOVectEntry*)ctxt->data_post;
#if 0
std::cout << "Wait tag=" << tag << ", serial=" << serial << " OK";
for (int j=0; j<ctxt->count_recv; ++j)
{
  std::cout << "iov[" << j << "]= [";
  for (int k=0; k< iov_data[j].len; ++k)
    std::cout << (int)((char*)iov_data[j].data)[k] << ' '; 
  std::cout << "]" << std::endl;
}
#endif
      /* If I am not last node, send the bloc to the next node  */
      if (indextosend != index_src)
      {
        _access.unlock();
        /* Form the message and resent it */
        ACOM::WriteAndSignal<BcastPipeline1> ws (_net, this, 0);
        ws->pack(&i, sizeof(int));
        ws->pack(&tag, sizeof(int));
        for (size_t k=0; k< ctxt->count_recv; ++k)
          ws->put( &iov_data[k] ); 
        ws.post(channel);
#if 0
std::cout << "Resent serial=" << i << " to node=" << ranks[indextosend] << std::endl;
#endif
        _access.lock();
      }

      /* deallocate header ctxt + curr */
      deallocate_cc( ctxt );
    }

    
    /* here all packets for immediate data are arrivied
       post recv for all deffered data 
    */
    _access.unlock();

    /* Unpack the data and get iovect description */
    ACOM::ByteSeqReader datastream;
    datastream.init( &im_data_seq );
    ACOM::IStream fin ( &datastream, iov_recv + iovsz_h, iovsz_d );
    if (count != 0)
      fmt->read(fin, buffer, count);
    im_data_seq.destroy();

    /* free immediate data */
    for (size_t i=0; i< iovsz_h; ++i)
      _net->deallocate_data( iov_recv[i].data );

    /* post iovector splitted by bloc for the defered data */
    int save_serial = serial;
    postrecv_iov( serial, tag, iov_recv + iovsz_h, iovsz_d );

    /* relock */  
    _access.lock();

    /* wait all message from save_serial to serial-1 */
    for (int i = save_serial; i < serial; ++i)
    {
      ctxt = wait_iov( i, tag );
      ACOM::IOVectEntry* iov_data = (ACOM::IOVectEntry*)ctxt->data_post;
#if 0
std::cout << "Wait tag=" << tag << ", serial=" << serial << " OK";
for (int j=0; j<ctxt->count; ++j)
{
  std::cout << "iov[" << j << "]= [";
  for (int k=0; k< iov_data[j].len; ++k)
    std::cout << (int)((char*)iov_data[j].data)[k] << ' '; 
  std::cout << "]" << std::endl;
}
#endif

      /* If I am not last node, send the bloc to the next node  */
      if (indextosend != index_src)
      {
        _access.unlock();
        /* Form the message and resent it */
        ACOM::WriteAndSignal<BcastPipeline1> ws (_net, this, 0);
        ws->pack(&i, sizeof(int));
        ws->pack(&tag, sizeof(int));
        for (size_t k=0; k< ctxt->count_recv; ++k)
          ws->put( &iov_data[k] ); 
        ws.post(channel);
#if 0
std::cout << "Resent serial=" << i << " to node=" << ranks[indextosend] << std::endl;
#endif
        _access.lock();
      }
      
      /* deallocate header ctxt + curr */
      deallocate_cc( ctxt );
    }
  } 
  return 0;
}


// -----------------------------------------------------------------------------
/** */
void BcastPipeline1::wait_bcast(
  handler wc
)
{
  if (wc ==0) return; // master
}


// -----------------------------------------------------------------------------
/** */
void BcastPipeline1::async_bcast(
    void*&                     buffer, 
    size_t&                    count,
    const ACOM::Format*        fmt,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_src, 
    int                        tag,
    void                       (*cb_func)(void*, void*),
    void*                      cb_arg1,
    void*                      cb_arg2
  )
{
  post_bcast( 
    buffer, 
    count,
    fmt,
    sz_ranks, 
    ranks, 
    index_self, 
    index_src, 
    tag
  );
  if (cb_func !=0) {
    (*cb_func)(cb_arg1, cb_arg2);
  }
}


}
