// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_dns.cpp,v $
// Copyright(c)'2001 by Apache Team
// Authors: Thierry Gautier
// Current implementation is based on a flat DNS structure : node 0 stores
// all informations about all other nodes.
//
// ==========================================================================
#include<ics_dns.h>

namespace ICS {

// --------------------------------------------------------------------------
/* operation code : 
*/
enum DNSOperation {
  OP_VOID   = 0, // no op
  OP_BIND   = 1, // bind
  OP_REBIND = 2, // rebind
  OP_RESOLV = 4, // resolv
  OP_UNBIND = 5, // unbind
  OP_REPLY  = 6  // reply to a request
};
  
/* exception code : 
*/
enum DNSException {
  E_OK =0,
  E_ALREADYBIND,
  E_BADNAME, 
  E_COMFAILURE,
  E_OTHER
};  

void throwit( DNSException ec ) throw( DNS::AlreadyBind, DNS::BadName )
{
  switch (ec) {
    case E_OK: 
      return;
    case E_ALREADYBIND: 
      ACOM::throwit( DNS::AlreadyBind() );
    case E_BADNAME: 
      ACOM::throwit( DNS::BadName() );
    case E_COMFAILURE: 
      ACOM::throwit( ACOM::ComFailure(ACOM::ComFailure::FAIL) );
    case E_OTHER: 
      ACOM::throwit( DNS::DnsError() );
  }
  return;
}


/** Data structure of each message :
*/
struct DNSMessage {
  DNSMessage() 
  : operation(OP_VOID), exceptioncode(0), 
    addr_flag(0), addr_code(0), addr_value(0),
    namingctxt(), name(), value()
  {}
  
  int         operation;
  int         exceptioncode;
  size_t      addr_flag;
  size_t      addr_code;
  size_t      addr_value;
  std::string namingctxt;
  std::string name;
  std::string value;
  
  static void pack  ( ACOM::Network::Request* m, const DNSMessage& c);
  static void unpack( ACOM::Network::Receipt* r, DNSMessage& c);
};




/* Services for DNS operation
*/
// --------------------------------------------------------------------------
class DNSRequest : public ACOM::ServiceAM {
public:
  void initialize( DNS* dns )
  { _dns = dns; }
protected:
  void call(
      ACOM::ComFailure::Code error_no,
      ACOM::Network::Receipt* receipt,
      ACOM::Network::Header* header,
      ACOM::IOVectEntry* vector,
      ACOM::IOVectSize count,
      void* ctxt
  ) throw(ACOM::ComFailure);
  DNS* _dns;
};



// --------------------------------------------------------------------------
DNS::DNS()
{
}


// --------------------------------------------------------------------------
DNS::~DNS()
{
}


// --------------------------------------------------------------------------
void DNS::initialize( ACOM::Network* net )
  throw (ACOM::InvalidArgument)
{
	if (net ==0) 
    ACOM::throwit(ACOM::InvalidArgument("nul argument"));
  if (net->get_property("primary") != "true")
    ACOM::throwit(ACOM::InvalidArgument("not a primary network"));
  _net = net;
  _access.init();
  _waitreply.init();
  
  _service   = new DNSRequest;
  _service->initialize( this );
  ACOM::Service::Manager.bind( ACOM::ServiceManager::ICS_LNS, ServiceNumber::ICS_DNS_REQUEST, _service );
  
}


// --------------------------------------------------------------------------
void DNS::destroy()
{
  delete _service;
  _access.destroy();
  _waitreply.destroy();
}
  

// --------------------------------------------------------------------------
void DNS::bind( const std::string& namectxt, const std::string& name, const std::string& value )
  throw (AlreadyBind, ACOM::ComFailure)
{
  if (_net->get_rank() ==0) {
    PTH::CriticalGuard critical(_access);
    NamingContext nc = _world[namectxt];
    
    /* look if name exist in nc */
    if (nc.find(name) != nc.end())
      ACOM::throwit( AlreadyBind() );
    
    /* add it in nc */
    nc[name] = value;
  }
  else { 
    /* To wait continuation */
    bool isrecv = false;
    int  exceptioncode = 0;

    /* Send a message to master */
    ACOM::Network::OutChannel* channel = _net->get_default_route( 0 );
    ACOM::AMMessage msg( _net, _service, 0 );
    DNSMessage m;
    m.operation  = OP_BIND;
    m.exceptioncode = 0;
    m.namingctxt = namectxt;
    m.name       = name;
    m.value      = value;
    m.addr_flag  = (size_t) &isrecv;
    m.addr_code  = (size_t) &exceptioncode;
    DNSMessage::pack( msg, m);
    
    PTH::CriticalGuard critical(_access);
    msg.post( channel );
    while (!isrecv) 
      _waitreply.wait( _access );
    throwit( (DNSException)exceptioncode );
  }
}  


// --------------------------------------------------------------------------
void DNS::rebind( const std::string& namectxt, const std::string& name, const std::string& value )
  throw (ACOM::ComFailure)
{
  if (_net->get_rank() ==0) {
    PTH::CriticalGuard critical(_access);
    NamingContext nc = _world[namectxt];
    
    /* add it in nc */
    nc[name] = value;
  }
  else { 
    /* To wait continuation */
    bool isrecv = false;
    int  exceptioncode = 0;

    /* Send a message to master */
    ACOM::Network::OutChannel* channel = _net->get_default_route( 0 );
    ACOM::AMMessage msg( _net, _service, 0 );
    DNSMessage m;
    m.operation  = OP_REBIND;
    m.exceptioncode = 0;
    m.namingctxt = namectxt;
    m.name       = name;
    m.value      = value;
    m.addr_flag  = (size_t) &isrecv;
    m.addr_code  = (size_t) &exceptioncode;
    DNSMessage::pack( msg, m);
    
    PTH::CriticalGuard critical(_access);
    msg.post( channel );
    while (!isrecv) 
      _waitreply.wait( _access );
    throwit( (DNSException)exceptioncode );
  }
}


// --------------------------------------------------------------------------
std::string DNS::resolve( const std::string& namectxt, const std::string& name )
  throw (BadName, ACOM::ComFailure)
{
  if (_net->get_rank() ==0) {
    PTH::CriticalGuard critical(_access);
    
    /* look is nc exist */
    if (_world.find(namectxt) == _world.end())
      ACOM::throwit( BadName() );

    NamingContext nc = _world[namectxt];
    
    /* look if name exist in nc */
    if (nc.find(name) == nc.end())
      ACOM::throwit( BadName() );
    
    /* return its value */
    return nc[name];
  }
  else { 
    /* To wait continuation */
    bool isrecv = false;
    int  exceptioncode = 0;
    std::string value  = "";

    /* Send a message to master */
    ACOM::Network::OutChannel* channel = _net->get_default_route( 0 );
    ACOM::AMMessage msg( _net, _service, 0 );
    DNSMessage m;
    m.operation  = OP_RESOLV;
    m.exceptioncode = 0;
    m.namingctxt = namectxt;
    m.name       = name;
    m.value      = "";
    m.addr_flag  = (size_t) &isrecv;
    m.addr_code  = (size_t) &exceptioncode;
    m.addr_value = (size_t) &value;
    DNSMessage::pack( msg, m);
    
    PTH::CriticalGuard critical(_access);
    msg.post( channel );
    while (!isrecv) 
      _waitreply.wait( _access );
    throwit( (DNSException)exceptioncode );
    return value;
  }
}  


// --------------------------------------------------------------------------
void DNS::unbind( const std::string& namectxt, const std::string& name )
  throw (BadName, ACOM::ComFailure)
{
  if (_net->get_rank() ==0) {
    PTH::CriticalGuard critical(_access);
    
    /* look is nc exist */
    if (_world.find(namectxt) == _world.end())
      ACOM::throwit( BadName() );

    NamingContext nc = _world[namectxt];
    
    /* look if name exist in nc */
    if (nc.find(name) == nc.end())
      ACOM::throwit( BadName() );
    
    /* suppress entry */
    nc.erase(name);
    return;
    
  } else {
    /* To wait continuation */
    bool isrecv = false;
    int  exceptioncode = 0;

    /* Send a message to master */
    ACOM::Network::OutChannel* channel = _net->get_default_route( 0 );
    ACOM::AMMessage msg( _net, _service, 0 );
    DNSMessage m;
    m.operation  = OP_UNBIND;
    m.exceptioncode = 0;
    m.namingctxt = namectxt;
    m.name       = name;
    m.value      = "";
    m.addr_flag  = (size_t) &isrecv;
    m.addr_code  = (size_t) &exceptioncode;
    DNSMessage::pack( msg, m);
  
    PTH::CriticalGuard critical(_access);
    msg.post( channel );
    while (!isrecv) 
      _waitreply.wait( _access );
    throwit( (DNSException)exceptioncode );
  }
}


// --------------------------------------------------------------------------
void DNSMessage::pack  ( ACOM::Network::Request* m, const DNSMessage& c )
{
  m->pack( &c.operation, sizeof(int) );
  m->pack( &c.addr_flag, sizeof(size_t) );
  m->pack( &c.addr_code, sizeof(size_t) );
  m->pack( &c.addr_value,sizeof(size_t) );
  { /* first string */
    size_t sz = c.namingctxt.size();
    m->pack( &sz, sizeof(size_t) );
    m->pack( c.namingctxt.c_str(), sz );
  }
  { /* second string */
    size_t sz = c.name.size();
    m->pack( &sz, sizeof(size_t) );
    m->pack( c.name.c_str(), sz );
  }
  { /* third string */
    size_t sz = c.value.size();
    m->pack( &sz, sizeof(size_t) );
    m->pack( c.value.c_str(), sz );
  }
}  


// --------------------------------------------------------------------------
void DNSMessage::unpack( ACOM::Network::Receipt* r, DNSMessage& c )
{
  r->unpack( &c.operation, sizeof(int) );
  r->unpack( &c.addr_flag, sizeof(size_t) );
  r->unpack( &c.addr_code, sizeof(size_t) );
  r->unpack( &c.addr_value,sizeof(size_t) );
  { /* first string */
    size_t sz = 0;
    r->unpack( &sz, sizeof(size_t) );
    char* s = (char*)alloca( sz+1 );
    r->unpack( s, sz );
    s[sz] =0;
    c.namingctxt = std::string( s );
  }
  { /* second string */
    size_t sz = 0;
    r->unpack( &sz, sizeof(size_t) );
    char* s = (char*)alloca( sz+1 );
    r->unpack( s, sz );
    s[sz] =0;
    c.name = std::string( s );
  }
  { /* third string */
    size_t sz = 0;
    r->unpack( &sz, sizeof(size_t) );
    char* s = (char*)alloca( sz+1 );
    r->unpack( s, sz );
    s[sz] =0;
    c.value = std::string( s );
  }
}


// --------------------------------------------------------------------------
void DNSRequest::call(
    ACOM::ComFailure::Code error_no,
    ACOM::Network::Receipt* receipt,
    ACOM::Network::Header* header,
    ACOM::IOVectEntry* vector,
    ACOM::IOVectSize count,
    void* ctxt
) throw(ACOM::ComFailure)
{
  DNSMessage m;
  DNSMessage::unpack( receipt, m );
  try {
    switch (m.operation) {
      case OP_VOID: break;
      case OP_BIND: 
          _dns->bind( m.namingctxt, m.name, m.value );
          m.value = "";
      break;
      case OP_REBIND: 
          _dns->rebind( m.namingctxt, m.name, m.value );
          m.value = "";
      break;
      case OP_RESOLV: 
          m.value = _dns->resolve( m.namingctxt, m.name );
      break;
      case OP_UNBIND: 
          _dns->unbind( m.namingctxt, m.name );
      break;
      case OP_REPLY: {
        /* recopy value if needed (OP_CODE is OP_RESOLV)
           and signal waiters
        */
        bool* flag = (bool*)m.addr_flag;
        std::string* dest = (std::string*)m.addr_value;
        int* ec = (int*)m.addr_code;
        if (dest !=0) 
          *dest = m.value;
        if (ec !=0) 
          *ec = m.exceptioncode;
        if (flag !=0) 
          *flag = true;
        _dns->_waitreply.broadcast();
        return; /* do not resend a message ! */
      }
      break;
      default:
        ACOM::throwit( DNS::DnsError() );
    }
  } catch (const DNS::AlreadyBind& ex) {
    m.exceptioncode = E_ALREADYBIND;
    m.value = "";
  } catch (const DNS::BadName& ex) {
    m.exceptioncode = E_BADNAME;
    m.value = "";
  } catch (const ACOM::ComFailure& ex) {
    m.exceptioncode = E_COMFAILURE;
    m.value = "";
  } catch (...) {
    m.exceptioncode = E_OTHER;
    m.value = "";
  }
  m.namingctxt = "";
  m.name = "";
  m.operation = OP_REPLY;

  ACOM::Network::OutChannel* channel = _dns->_net->get_default_route( header->src );
  ACOM::AMMessage msg( _dns->_net, _dns->_service, 0 );
  DNSMessage::pack( msg, m);  
  msg.post( channel );
}

} // namespace ICS
