// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_lighthashmap.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation light hash map class specialized for some services.
// For each entry, the list could be managed using a priority order,
// by using insert_front/_back.
// $Id: ics_lighthashmap.h,v 1.2 2003/07/15 17:22:44 thierry Exp $
//
// ==========================================================================
#ifndef _INUK_HASH_MAP_H_
#define _INUK_HASH_MAP_H_
#include <iostream>
#include <ics_conf.h>

namespace ICS {

/*
 * HashMap to retreive a rendez-vous
 * KEY should have operators:
 * - bool operator==( const KEY&, const KEY& ): comparaison of two keys
 * - int  operator%( const KEY&, int size ): should return a hashcode
 * less than size.
 */
template<class KEY, class VALUE>
class LHashMap {
public:
  class entry  { 
    friend class LHashMap;
    entry*       _next;
    entry*       _prev;
    long         _hashkey;
  public:
    KEY          first;
    VALUE        second;
  };

  /* cstor */
  LHashMap( int size =256, size_t* ca =0, size_t* cf =0 );

  /* dcstor */
  ~LHashMap();

  /* iterator */
  typedef entry* iterator;

  /* return the size */
  size_t size( ) const; 

  /* find */
  iterator find( KEY key ); 

  /* insert */
  iterator insert( KEY key, VALUE val ); 

  /* erase */
  void erase( iterator i ); 

  /* reinsert: put i from a hash table to an other one */
  void reinsert( LHashMap<KEY,VALUE>& owner, iterator i ); 

protected:
  int _HASH_MAP_SIZE;
  entry** _head;
  entry** _tail;
  size_t  _size;
  entry*  _freelist;
  size_t  _cached;
  size_t* _count_alloc;
  size_t* _count_free;
};


/* 
 * inline definition
 */
template<class KEY, class VALUE>
inline LHashMap<KEY,VALUE>::LHashMap(int size, size_t* ca, size_t* cf )
  : _HASH_MAP_SIZE(size), _head(0), _tail(0), _size(0), 
    _freelist(0), _cached(0),
    _count_alloc(ca), _count_free(cf)
{
  _head = new entry*[size];
  _tail = new entry*[size];
  for (int i=0; i<_HASH_MAP_SIZE; ++i)
    _head[i] = _tail[i] = 0;
}


template<class KEY, class VALUE>
inline LHashMap<KEY,VALUE>::~LHashMap()
{
  for (int i=0; i<_HASH_MAP_SIZE; ++i)
  {
    if (_head[i] !=0) {
      entry* curr = _head[i];
      while (curr !=0) {
        entry* next = curr->_next;
        delete curr;
        --_size;
        if (_count_free !=0) ++*_count_free;
        curr = next;
      }
      _head[i] = _tail[i] = 0;
    }
  }
  entry* curr = _freelist;
  while (curr !=0) {
    entry* next = curr->_next;
    delete curr;
    if (_count_free !=0) ++*_count_free;
    --_cached;
    curr = next;
  }
  delete [] _head;
  delete [] _tail;
}


template<class KEY, class VALUE>
inline size_t LHashMap<KEY,VALUE>::size( ) const
{ return _size; }



#define INUK_HASH_VALUE(key) ( key % (_HASH_MAP_SIZE) )

template<class KEY, class VALUE>
inline typename LHashMap<KEY,VALUE>::iterator LHashMap<KEY,VALUE>::find( KEY key )
{
  int index = INUK_HASH_VALUE(key);
  if (_head[index] ==0) return 0;
  entry* curr = _head[index];
  do {
    if (curr->first == key) return curr;
    curr = curr->_next;
  } while (curr !=0);
  return 0;
}


/* insert */
template<class KEY, class VALUE>
typename LHashMap<KEY,VALUE>::iterator LHashMap<KEY,VALUE>::insert( KEY key, VALUE val )
{
  int index = INUK_HASH_VALUE(key);

  /* new entry in the hash table */
  entry* newitem = 0;
  if (_freelist ==0) {
    newitem = new entry;
    if (_count_alloc !=0) ++*_count_alloc;
  } else {
    newitem = _freelist; 
    _freelist = newitem->_next;
    --_cached;
  }
  newitem->_hashkey = index;
  newitem->first = key;
  newitem->second = val;

  /* add at the end */
  newitem->_prev = _tail[index];
  newitem->_next = 0;
  if (_tail[index] ==0) {
    _head[index] = _tail[index] = newitem;
  } else {
    _tail[index]->_next = newitem;
    _tail[index] = newitem;
  }
  ++_size;
  return newitem;
}



template<class KEY, class VALUE>
inline void LHashMap<KEY,VALUE>::erase( iterator i )
{
  if (i ==0) return;
  --_size;
#ifdef ICS_DEBUG
  if (_size > 1000) throw;
#endif  

  if (i->_next !=0) {
    i->_next->_prev = i->_prev;
  } else {
    _tail[i->_hashkey] = i->_prev;
  }
  if (i->_prev !=0) {
    i->_prev->_next = i->_next;
  } else {
    _head[i->_hashkey] = i->_next;
  }
  if (_cached > 32) {
    delete i;
    if (_count_free !=0) ++*_count_free;
  } else {
    i->_next = _freelist;
    i->_prev =0;
    i->_hashkey = -1;
    _freelist = i;
    ++_cached;
  }
}


template<class KEY, class VALUE>
inline void LHashMap<KEY,VALUE>::reinsert( LHashMap<KEY,VALUE>& owner, iterator i )
{
  if (i ==0) return;
  if (i->_next !=0) {
    i->_next->_prev = i->_prev;
  } else {
    owner._tail[i->_hashkey] = i->_prev;
  }
  if (i->_prev !=0) {
    i->_prev->_next = i->_next;
  } else {
    owner._head[i->_hashkey] = i->_next;
  }
  --owner._size;
#ifdef ICS_DEBUG
  if (owner._size > 1000) throw;
#endif
  
  int newhashkey = INUK_HASH_VALUE(i->first);
  i->_hashkey = newhashkey;

  /* add at the end */
  i->_prev = _tail[newhashkey];
  i->_next = 0;
  if (_tail[newhashkey] ==0) {
    _head[newhashkey] = _tail[newhashkey] = i;
  } else {
    _tail[newhashkey]->_next = i;
    _tail[newhashkey] = i;
  }
  ++_size;
}

}

#endif
