// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_network.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Base class for network with collective operation
// Please read node in .cpp file.
// $Id: ics_network.h,v 1.16 2003/07/29 20:01:54 thierry Exp $
//
// ==========================================================================
#ifndef _INUK_NETWORK_H_
#define _INUK_NETWORK_H_
#include <acom>
#include <ics_conf.h>
#include <ics_lighthashmap.h>

namespace ICS {

/* fwd declaration */
class Network;

/* fwd declaration */
class ThreadRedirec;

/* fwd declaration */
class QueueWork;

using ACOM::uint16_t;
using ACOM::uint32_t;
using ACOM::uint8_t;



// ---------------------------------------------------------------------------
struct RedirectServiceCallBack : public ACOM::Callback {
  /* */
  RedirectServiceCallBack( ACOM::Network::Request* req,    /* request to notify */
                           ACOM::Network::CallContext* cc  /* cc to release */
                         ) 
    : _req(req), _call_cc(cc) {}

  /* release call cc */
  void done( ACOM::Network::Request* );

private:
  ACOM::Network::Request*     _req;
  ACOM::Network::CallContext* _call_cc;
};



// -----------------------------------------------------------------------------
/** RedirectService.
 * Such a service is used to redirect a remote invocation to internal service
 * to an invocation to a user's level service. It is used to mask added data.
 */
class RedirectService : public ACOM::ServiceCC {
public:
  /** initialize and attach a network 
  */
  virtual void initialize( Network* net );

  /** Return attached network
  */
  Network* get_net() 
  { return _attachnet; }

protected:
  /** Main server entry point for call extract
  */
  void server_call( ACOM::Network::CallContext* call_cc ) 
      throw (ACOM::ComFailure);

  /** Main server entry point for get_iovect
  */
  void server_get_iovect( ACOM::Network::CallContext* call_cc )
      throw (ACOM::ComFailure);  
  
protected:
  Network* _attachnet;
};




// -----------------------------------------------------------------------------
/** Network with collective communication capabilities.
 Such network relies on an existing network to handle point 2 point communication.
 It should be declared and registered as other standard network in Inuktitut.
 Such a network should be initialized on top of an existing network.
 Bi-point channels are always accessible to the user using the standard network interface
 (get_default_route).
 ICS::Network extends the method of a base Inuktitut's ACOM::Network by providing
 a way to declare channels for collective operations. 
 The allocation of data and request are handle by the underlaying network.
 A Channel for collective operations has a name which is a tag (integer). 
 Such a channel corresponds to a set of nodes, each node is given by its ranks on the 
 underlaying network. 
 All channels should have a different name. But name are only valid for a network : 
 because the user could instanciate multiple network with collective communication 
 capabilites, same name can exists for each such network.
 
 A ICS::Network class is functional "as it" : it could be initialized on top of
 an existing network. Its name is ICS_<networkname>. Such simple ICS network
 uses multiple threads to handle communication (both input and redirection).
 
 Current limitations are :
 - name of collective channel should be stored on 32 bits
*/
class Network: public ACOM::Network {
public:
  // --------------------------------------------------------------------------
  /** Constructor
  */
  Network();

  /** Destructor
  */
  ~Network();

  // ---------------------------------------------------------------------------
  /** extended header needed by all collective communications
   */
  struct Header  {
    Header() 
     : saveid(0), tag(0), netid(0)
    {}
    uint32_t     saveid;      // - save id of original header id
    uint32_t     tag;         // - name of the channel
    uint16_t     netid;       // - name of the network instance
  };
  

  // ---------------------------------------------------------------------------
  /** extended classes for broadcast purpose.
  */
  class OutChannel: public ACOM::Network::OutChannel {
  public:
    /** Constructor
    */
    OutChannel( );

    /** destructor
    */
    ~OutChannel();
    
    /** Initialize the collective channel.
        See in derived classes other attributs that could be passed to a channel.
        @param net (IN) the associated network to this channel
        @param tag (IN) the name of the channel, an error is thrown if the name already exists
    */
    void init( Network* net, uint32_t tag );

    /** Destroy an existing channel.
        After the invocation to this method, the channel has no more existance on this node.
        Any pending or new communication that refer the name of the destroy channel will
        invoke a communication error.
    */
    virtual void destroy( );

    /** Redirect to a call to the method post on the sub network or
        call to BcastNetwork::process depending of the channel.
    */
    void post( ACOM::Network::Request* r ) throw (ACOM::ComFailure);

  public: /* to all derived service of forwarding */
    uint32_t   _tag;           /* name of the channel */
  };


  // ---------------------------------------------------------------------------
  /* inherited and redefined methods:
  */
  int initialize( int* argc, char** argv[] )
    throw (ACOM::NetworkError);

  int commit () throw (ACOM::NetworkError);

  int terminate () throw (ACOM::NetworkError);

  ACOM::Network::Request* allocate_request () throw (ACOM::BadAlloc);

  void deallocate_request( ACOM::Network::Request* req ) throw (ACOM::InvalidArgument);

  void* allocate_data( size_t sz )
    throw (ACOM::BadAlloc);

  void deallocate_data(  void* data )
    throw (ACOM::InvalidArgument);

  void close_channel( ACOM::Network::OutChannel* ch )
      throw(ACOM::ComFailure);

  /** Redirect call to process 
  */
  void post( ACOM::Network::Request* req, ACOM::Network::OutChannel* ch)
    throw (ACOM::ComFailure);
  
public:
  // ---------------------------------------------------------------------------
  /* inherited methods redefined to use those of the underlaying network.
  */
  ACOM::Network::OutChannel* open_channel( LocalNodeNo peer )
       throw(ACOM::ComFailure)
  { return _subnet->open_channel(peer); }    


  /** */
  void barrier()
    throw (ACOM::ComFailure)
  { _subnet->barrier(); }


  /** */
  ACOM::Network::OutChannel* get_default_route( LocalNodeNo peer )
       throw (ACOM::NetworkError)
  { return _subnet->get_default_route(peer); }


  /** Allocate a new call context to treats incomming message
      The method MUST no has a default implementation.
  */
  CallContext* allocate_callcontext(size_t sz) throw (ACOM::BadAlloc);

  /** Deallocate a call context
      The method MUST no has a default implementation.
  */
  void deallocate_callcontext(CallContext*) throw (ACOM::InvalidArgument);

  /** Reallocate of the stack of the call context
  */
  ACOM::Network::CallContext* resize_callcontext(ACOM::Network::CallContext* ctxt, size_t sz ) throw (ACOM::BadAlloc);

  /** Acquire a call context for a new incomming communication
  */
  ACOM::Network::CallContext* acquire_callcontext() throw (ACOM::BadAlloc);
  
  /** Release a call context 
  */
  void release_callcontext(ACOM::Network::CallContext*) throw (ACOM::InvalidArgument);

  // ---------------------------------------------------------------------------
  /* extended methods for collective communication purpose.
  */
  /* derived class should implements in order to set the name of the
     network. This method must be called to register the name in the
     data base. 
  */
  int initialize( int* argc, char** argv[], ACOM::Network* netInit ) 
    throw (ACOM::NetworkError);


  /** Called to process a request or a incomming message.
      The purpose of this method is to code the algorithm to broadcast or reduce
      message. Derived class should redefined the method. Two calling conventions
      are possible to call the method: from a post (see broadcast network), in this
      case the request argument is non null and the call context could be or could
      not be null (for broadcast network it is null, not for reduce network); 
      from a incomming message, in this case the request argument is null. 
      The implementation of this method must release the call_cc.
      @param the channel used for collective communication
      @param call_cc (IN) the callcontext to be processed
      @param req (IN) the request to be processed
  */
  virtual void process( ACOM::Network::OutChannel* ch, 
                        ACOM::Network::CallContext* call_cc, 
                        ACOM::Network::Request* r
                      )
    throw (ACOM::ComFailure)  =0;

  /** Return the broadcast channel associated to the tag 'tag'
     @param tag(IN) the name of the broadcast channel
     @retval a pointer to the broadcast channel or 0
  */
  OutChannel* get_channel( uint32_t tag );
  
  /** Return the underlaying network object
  */
  ACOM::Network* get_subnet()
  { return _subnet; }

  /** Return reference to the lock
  */
  PTH::Mutex& mutex()
  { return _access; }

  /* structure to shared work between consummer (the thread)
     and producer (service)
  */
  struct Work : PTH::Queable<Work> {
    ACOM::Network*              subnet; /* use either net or subnet, only one of them is !=0 */
    Network*                    net; 
    ACOM::Network::OutChannel*  ch;
    ACOM::Network::CallContext* cc;
    ACOM::Network::Request*     req;
  };

protected:
  /* -------------------------------------------------------------------------------- */
  /* -------------------------------------------------------------------------------- */
  ACOM::Network*                  _subnet;           // underlaying network
  PTH::Mutex                      _access;           // lock access to internal data
  PTH::Condition                  _cond;             // lock access to internal data
  RedirectService*                _service;          // service for redirection...
  LHashMap<uint32_t, OutChannel*> _allchannels;      // list of all channels openned

private:
  friend class ThreadRedirec;
  friend class RedirectService;
  friend class OutChannel;
};
 


// -----------------------------------------------------------------------------
/* Basic Queue of works : FIFO ordering
*/
class BasicQueueWork {
public:
  /** build and initialize a new queue
  */
  BasicQueueWork();

  /** destroy the queue
  */
  ~BasicQueueWork();
  
  /** Pop in front a new work.
      If the queue is empty then the caller is blocked until a new
      piece of work is pushed. On return value is either a pointer
      to a work that should be freed by the thread, or 0. This later
      case occurs when the blocked thread is wakeup by an other which
      makes a call to clear or wake_up on the queue.
      @return a pointer to work or 0
  */
  Network::Work* pop( );

  /** push in back in the queue a new piece of work
  */
  void push( Network::Work* new_work );

  /** wake up all waiting threads.
      All threads block on pop, have a return value 0.
  */
  void wake_up();

  /** clear the queue and wake up all waiting threads.
      All threads block on pop, have a return value 0.
  */
  void clear();

  /** return the size of the list
  */
  size_t size();
protected:
  PTH::Condition            _cond;
  PTH::Mutex                _access;  
  int                       _nwaiters;
  bool                      _wakeup;
  size_t                    _size;
  PTH::Queue<Network::Work> _queue;
};



// -----------------------------------------------------------------------------
/* Queue of works as a list of queue managed on round robin fashion
*/
class QueueWork {
public:
  /** build and initialize a new queue
  */
  QueueWork();

  /** destroy the queue
  */
  ~QueueWork();
  
  /** Return the number of basic queue
  */
  int size();
  
  /** Add a basic queue
  */
  void add_queue(BasicQueueWork* q);
  
  /** Pop a new work.
      If the queue is empty then the caller is blocked until a new
      piece of work is pushed. On return value is either a pointer
      to a work that should be freed by the thread, or 0. This later
      case occurs when the blocked thread is wakeup by an other which
      makes a call to clear or wake_up on the queue.
      @param index the associated basic queue for pop
      @return a pointer to work or 0
  */
  Network::Work* pop( int index );

  /** push in the queue a new piece of work
  */
  void push( Network::Work* new_work );

  /** wake up all waiting threads.
      All threads block on pop, have a return value 0.
      @param bq the basicqueue on which threads should be wakeup, -1 : all threads
  */
  void wake_up( int bq = -1 );

  /** clear the queue and wake up all waiting threads.
      All threads block on pop, have a return value 0.
  */
  void clear();

  /** Allocate a new work
  */
  Network::Work* allocate( );

  /** Deallocate a new work
  */
  void deallocate( Network::Work* w );

protected:
  friend class Ics;
  PTH::Mutex                    _access;  
  int                           _size;
  int                           _ntopush;
  std::vector<BasicQueueWork*>  _all_queues;
  PTH::StackList<Network::Work> _free_work;        // list of free work
};


// -----------------------------------------------------------------------------
/* thread that handles list of CallContext to redirect them to next node
*/
class ThreadRedirec : public PTH::ThreadModel<ThreadRedirec> {
public:
  /* Consumer of cc, entry point of the thread */
  void* run();
  
  /* Stop the thread */
  void stop();
  
  /* kill the thread */
  void kill();
  
  /* re start a stopped thread */
  void restart();

protected:
  /* build a thread that will redirect a message to dest nn 
  */
  ThreadRedirec( int index );
  
  /* destroy the thread objet */
  ~ThreadRedirec( );
  
public:
  static std::vector<ThreadRedirec*> allthreads;  /* list of all threads for redirection */
  static QueueWork*                  queue;       /* queue for all the work to redirect */

protected:
  PTH::Mutex                         _access;
  PTH::Condition                     _condstop;
  bool                               _stop;       /* true if the thread should stop */
  bool                               _kill;       /* true if the thread should exit */
  int                                _bq_index;   /* index to pop on a specific basic queue */
  friend class Network;
  friend class Ics;
};




/*
 * inline definitions
 */
inline size_t BasicQueueWork::size()
{ return _size; }

inline int QueueWork::size()
{ return _size; }

inline void QueueWork::add_queue(BasicQueueWork* q)
{ 
  PTH::CriticalGuard critical(_access);
  _all_queues.push_back(q); ++_size; 
}

inline Network::Work* QueueWork::pop( int index )
{ 
  ACOM_ASSERT_DEBUG( index >=0, "bad assertion");
  if (index >_size) index %= _size;
  return _all_queues[index]->pop(); 
}

inline void QueueWork::push( Network::Work* new_work )
{
  _all_queues[_ntopush]->push( new_work );
  if (++_ntopush >= _size) _ntopush = 0;
}

inline void QueueWork::wake_up( int bq )
{
  PTH::CriticalGuard critical(_access);
  if (bq >0) {
    ACOM_ASSERT_DEBUG( bq <_size, "bad assertion");
    _all_queues[bq]->wake_up();
  }
  else {
    for (int i=0; i<_size; ++i)
      _all_queues[i]->wake_up();
  }
}

inline void QueueWork::clear()
{
  PTH::CriticalGuard critical(_access);
  for (int i=0; i<_size; ++i)
    _all_queues[i]->wake_up();
  _all_queues.clear();
}
 
}

#endif
