// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_redops.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Definition of reduction operation
// $Id: ics_redops.h,v 1.2 2003/07/28 08:05:40 thierry Exp $
//
// ==========================================================================
#ifndef _INUK_REDUCEOPS_H_
#define _INUK_REDUCEOPS_H_
#include <acom>
#include <ics_conf.h>
#include <algorithm>


namespace ICS {

// -------------------------------------------------------------------------
/* For the cumul function
*/
class CumulOperatorService : public ACOM::ServiceCC {
public:
  /** Call to make reduction : (call_cc_inout, call_cc_in) -> call_cc_inout.
      If call_cc_in is 0 then call_cc_inout is the first received message
      before local post occurs.
      User could redefine server_get_iovect to implement address translation
      and allocation of data in user space.
  */
  virtual void server_reduce( ACOM::Network::CallContext* call_cc_inout, 
                              const ACOM::Network::CallContext* call_cc_in ) 
      throw (ACOM::ComFailure);

protected:
};

/*
  ICS cumul operator format is an object that store a pointer to a function
  for reduction. This is a format: the communication of the function is done
  by sending the id of the format and recover it by casting id to a concrete
  pointer to a cumul operator format object.
  
  Two kinds of operator are defined:
  - F1 == unary (1): can only be performed on one element with a given format 
  - FN == N-ary (n): can only be performed on a vector of elements with a 
    given format.
  Default operators permit to pass from F1 to FN:
  1/  F1(dest, fmt, val) = FN(dest, fmt, val, 1)
  2/  FN(dest, fmt, val, count) = F1(dest_i, fmt, val_i) for all elements of 
    the vector
  Note that fmt is the format of val !, the format of dest is known by the caller
  or reduce operations.
  We assume that these both operators could be defined. 
  They were called only on call to cumul !
*/

// -------------------------------------------------------------------------
/* For the cumul function
*/
class CumulOperatorFormat : public ACOM::Format {
public:
  /** Type for the cumul function: should be communicable
   */
  typedef void (CumulOperatorFormat::*CumulOperator1)( void* dest, const ACOM::Format* fmt, const void* val ) const;

  /** Type for the cumul function: should be communicable
   */
  typedef void (CumulOperatorFormat::*CumulOperatorN)( void* dest, const ACOM::Format* fmt, const void* val, size_t count ) const;

  /** The default cstor
   */
  CumulOperatorFormat( );

  /** The cstor for unary function
   */
  CumulOperatorFormat( CumulOperator1 fnc );

  /** The cstor for N-arity function
   */
  CumulOperatorFormat( CumulOperatorN fnc );

  /** The cstor for N-arity function
   */
  CumulOperatorFormat( CumulOperator1 fnc1, CumulOperatorN fncN );

  /**@name operator that acts on one element 
  */
  //@{
  /** Set the cumul function on this site
   */
  void set_cumul_fnc( CumulOperator1 fnc )
  { _cumulfnc1 = fnc; }

  /** Get the cumul function for this 
   */
  CumulOperator1 get_cumul_fnc1( ) const
  { return _cumulfnc1; }

  /** Apply the method
   */
  void cumul( void* dest, const ACOM::Format* fmt, const void* val ) const
  { (this->*_cumulfnc1)(dest, fmt, val); }

  /** Get the cumul function from an id 
   */
  static CumulOperator1 get_cumul_fnc1( int id )
  { return ((CumulOperatorFormat*)ACOM::Format::get_format(id))->_cumulfnc1; }

  /** Default accumulate function
  */
  void default_fnc1( void* dest, const ACOM::Format* fmt, const void* val ) const;
  //@}


  /**@name operator that acts on vector 
  */
  //@{
  /** Set the cumul function on this site
   */
  void set_cumul_fnc( CumulOperatorN fnc )
  { _cumulfncN = fnc; }

  /** Get the cumul function for this 
   */
  CumulOperatorN get_cumul_fncN( ) const
  { return _cumulfncN; }

  /** Apply the method
   */
  void cumul( void* dest, const ACOM::Format* fmt, const void* val, size_t count ) const
  { (this->*_cumulfncN)(dest, fmt, val, count); }

  /** Get the cumul function from an id 
   */
  static CumulOperatorN get_cumul_fncN( int id )
  { return ((CumulOperatorFormat*)ACOM::Format::get_format(id))->_cumulfncN; }

  /** Default accumulate function
  */
  void default_fncN( void* dest, const ACOM::Format* fmt, const void* val, size_t count ) const;
  //@}


  /** Misc
  */
  //@{
  /** Get the cumul function from an id 
   */
  static const CumulOperatorFormat* get_cumul_format( int id )
  { return (CumulOperatorFormat*)ACOM::Format::get_format(id); }

  static const CumulOperatorFormat& Null;
  //@}

protected:
  CumulOperator1 _cumulfnc1;
  CumulOperatorN _cumulfncN;
};



// --------------------------------------------------------------------
/* Elementary operation
*/
template<class T>
struct UnitaryOpAdd {
  void operator()( T& res, const T& v)
  { res += v; }
};
template<class T>
struct UnitaryOpSub {
  void operator()( T& res, const T& v)
  { res -= v; }
};
template<class T>
struct UnitaryOpMul {
  void operator()( T& res, const T& v)
  { res *= v; }
};
template<class T>
struct UnitaryOpDiv {
  void operator()( T& res, const T& v)
  { res /= v; }
};
template<class T>
struct UnitaryOpMax {
  void operator()( T& res, const T& v)
  {  res = std::max(res,v); }
};
template<class T>
struct UnitaryOpMin {
  void operator()( T& res, const T& v)
  {  res = std::min(res,v); }
};


// --------------------------------------------------------------------
/* Vector operation
*/
template<class T, class OP>
class VectorOp : public CumulOperatorFormat {
  static OP op;
public:
  void def_cumul1( void* dest, const ACOM::Format* fmt, const void* val ) const
  {
    T* d = (T*)dest; const T* v = (const T*)val;
#if 0
    std::cout << "d=[ " << d[0] << "] " << std::endl;
    std::cout << "v=[ " << v[0] << "] " << std::endl;
#endif
    op(*d, *v);
#if 0
    std::cout << "d=[ " << d[0] << "] " << std::endl;
#endif
  }
  void def_cumulN( void* dest, const ACOM::Format* fmt, const void* val, size_t count ) const
  {
    T* d = (T*)dest; const T* v = (const T*)val;

#if 0
    std::cout << "d=[ ";
    for (size_t i=0; i<count; ++i) {
      std::cout << d[i] << ' ';
    }
    std::cout << "] " << std::endl;
    
    std::cout << "v=[ ";
    for (size_t i=0; i<count; ++i) {
      std::cout << v[i] << ' ';
    }
    std::cout << "] " << std::endl;
#endif

    for (size_t i=0; i<count; ++i) {
      op(d[i], v[i]);
    }

#if 0
    std::cout << "d=[ ";
    for (size_t i=0; i<count; ++i) {
      std::cout << d[i] << ' ';
    }
    std::cout << "] " << std::endl << std::endl << std::endl;
#endif
  }
  static const CumulOperatorFormat theformat;
  static const CumulOperatorFormat* const format;
  static const CumulOperatorFormat* get_format();
};

template<class T, class OP>
OP VectorOp<T,OP>::op;

template<class T, class OP>
const CumulOperatorFormat VectorOp<T,OP>::theformat(
       (CumulOperatorFormat::CumulOperator1)&VectorOp<T,OP>::def_cumul1,
       (CumulOperatorFormat::CumulOperatorN)&VectorOp<T,OP>::def_cumulN
     );

template<class T, class OP>
const CumulOperatorFormat* const VectorOp<T,OP>::format = &VectorOp<T,OP>::theformat;

template<class T, class OP>
inline const CumulOperatorFormat* VectorOp<T,OP>::get_format()
{ return VectorOp<T,OP>::format; }


/* Predefined operation */
template<class T>
struct OpAdd : public VectorOp<T,UnitaryOpAdd<T> >
{ };
template<class T>
struct OpSub : public VectorOp<T,UnitaryOpSub<T> >
{ };
template<class T>
struct OpMul : public VectorOp<T,UnitaryOpMul<T> >
{ };
template<class T>
struct OpDiv : public VectorOp<T,UnitaryOpDiv<T> >
{ };
template<class T>
struct OpMax : public VectorOp<T,UnitaryOpMax<T> >
{ };
template<class T>
struct OpMin : public VectorOp<T,UnitaryOpMin<T> >
{ };


// --------------------------------------------------------------------
/* Scan operation: val is a vector, dest is an element
*/
template<class T, class OP>
class ScanOp : public CumulOperatorFormat {
  static OP op;
public:
  void def_cumul1( void* dest, const ACOM::Format* fmt, const void* val ) const
  {
    T* d = (T*)dest; const T* v = (const T*)val;
    op(*d, *v);
  }
  void def_cumulN( void* dest, const ACOM::Format* fmt, const void* val, size_t count ) const
  {
    T* d = (T*)dest; const T* v = (const T*)val;
    for (size_t i=0; i<count; ++i)
      op(*d, v[i]);
  }
  static const CumulOperatorFormat theformat;
  static const CumulOperatorFormat* const format;
  static const CumulOperatorFormat* get_format();
};

template<class T, class OP>
OP ScanOp<T,OP>::op;

template<class T, class OP>
const CumulOperatorFormat ScanOp<T,OP>::theformat(
       (CumulOperatorFormat::CumulOperator1)&ScanOp<T,OP>::def_cumul1,
       (CumulOperatorFormat::CumulOperatorN)&ScanOp<T,OP>::def_cumulN
     );

template<class T, class OP>
const CumulOperatorFormat* const ScanOp<T,OP>::format = &ScanOp<T,OP>::theformat;

template<class T, class OP>
inline const CumulOperatorFormat* ScanOp<T,OP>::get_format()
{ return ScanOp<T,OP>::format; }



/*
 * inline definitions
 */

 
}

#endif
