// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_reduce_network.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of virtual base class for all reduce algorithms
// Please read node in .cpp file.
// $Id: ics_reduce_network.h,v 1.10 2003/07/28 08:05:41 thierry Exp $
//
// ==========================================================================
#ifndef _INUK_REDUCE_NETWORK_H_
#define _INUK_REDUCE_NETWORK_H_
#include <ics_network.h>
#include <ics_redops.h>


namespace ICS {

/* fwd declaration */
class ReduceNetwork;

// -----------------------------------------------------------------------------
/** OpReduceService : service that is called to reduce data
 */
typedef CumulOperatorService OpReduceService;


// -----------------------------------------------------------------------------
/** ReduceNetwork with reduce (a la MPI) capabilities.
 * Such network relies on an existing network to handle point 2 point communication.
 * Allocation of request and channel is specific to such network in order
 * to add extra information as immediate data. 
 In order to implement and algorithm for the broadcast collective operation, the
 derived class should defined a OutChannel that inherit from OutChannel. This 
 derivation must implement the OutChannel::post which remains undefined.
 The new derived network class should also implement :
  - virtual OutChannel* open_channel( int tag, int count_peer, LocalNodeNo* peers, 
                                      int index_self )
      throw(ACOM::ComFailure);
  - virtual void close_channel( OutChannel* ch )
      throw(ACOM::ComFailure);
  In order to allocate the right derived OutChannel.
 */
class ReduceNetwork: public Network {
public:
  // --------------------------------------------------------------------------
  /** Constructor
  */
  ReduceNetwork();

  /** Destructor
  */
  ~ReduceNetwork();

  // ---------------------------------------------------------------------------
  /** class for extended header
  */
  struct Header : public Network::Header {
    Header() 
     : Network::Header(), index_dest(0)
    {}
    uint32_t     opid;      /* id of reduction operator */
    uint32_t     index_dest; 
    uint32_t     key;       /* key for multifunction rendez-vous */
  };


  // ---------------------------------------------------------------------------
  /** extended classes for broadcast purpose.
  */
  class OutChannel: public Network::OutChannel {
  protected:
    /* Post should be redefined in derived class */
    //virtual void post( ACOM::Network::Request* r ) throw (ACOM::ComFailure);

    /** Constructor
    */
    OutChannel( );

    /** destructor
    */
    ~OutChannel();
    
  public:
    /** Initialize a channel for reduction.
        Such a channel reduce a value send by serveral nodes (peers) on a single
        node (peers[index_dest]). index_self is the index in the array peers of
        the rank of the current node. The service op is called on all nodes 
        that participate to the reduction. On the final node, this service is called
        before the call of the user service specified at the creation of the message.
        All nodes must send a message with the same user service. If the service
        for each of these messages differs, the behaviour is unspecified.
        @param net (in) the attached network for this channel
        @param tag (in) the name of the channel
        @param count_peer (in) the number of nodes that will participate to the reduction
        @param index_self (in) index in peers of the current node
        @param index_dest (in) index in peers of the final node that will handle the final data
    */
    virtual void init( ReduceNetwork* net, int tag, OpReduceService* op,
                       int count_peer, LocalNodeNo* peers, int index_self, int index_dest );

    virtual void destroy( );

  public: /* to all derived service of forwarding */
    OpReduceService* _op;          /* reduction operator */
    uint32_t         _serial;      /* to generate key for multi-function call */
    uint16_t         _count_peers; /* number of nodes involved */
    uint16_t*        _peers;       /* list of their ranks  */
    uint16_t         _index_self;  /* index of current node in _peers */
    uint16_t         _index_dest;  /* index of final node */
  };

  // ---------------------------------------------------------------------------
  /* inherited and redefined methods:
  */
  int initialize( int* argc, char** argv[], ACOM::Network* netInit ) 
    throw (ACOM::NetworkError);

  // ---------------------------------------------------------------------------
  /* extended methods for broadcast purpose.
  */
  /** Create and initialize a channel for reduction.
      @see init
      @param net (in) the attached network for this channel
      @param tag (in) the name of the channel
      @param count_peer (in) the number of nodes that will participate to the reduction
      @param index_self (in) index in peers of the current node
      @param index_dest (in) index in peers of the final node that will handle the final data
  */
  virtual OutChannel* open_channel( int tag, OpReduceService* op,
                                    int count_peer, LocalNodeNo* peers, int index_self, int index_dest )
       throw(ACOM::ComFailure) =0;

protected:
  friend class OutChannel;
  
  /* hash table of <tag,serial> to detect rendez vous
  */
  struct Key {
    /* */
    Key() 
     : tag(0), serial(0) 
    {}
    /* */
    Key(uint32_t t, uint32_t s) 
     : tag(t), serial(s) 
    {}
    /* used to compute the hash value */
    operator int() const 
    { return tag | serial; }
    /* used to compute the hash value */
    bool operator ==(const Key& k) const 
    { return (tag == k.tag) && (serial == k.serial); }

    /* used to compute the hash value */
    bool operator !=(const Key& k) const 
    { return (tag != k.tag) || (serial != k.serial); }
    uint32_t tag;
    uint32_t serial;
  };
  
  /* To detect rendez-vous
   */
  struct Rendezvous {
    /** Default constructor
    */
    Rendezvous() 
      : rcv_msg(0), wait_msg(0), call_cc(0), req(0)
    {}

    /** Return true is rendez vous occurs 
    */
    bool is_ok() 
    { return rcv_msg == wait_msg; }
    
    uint32_t                    rcv_msg; 
    uint32_t                    wait_msg;
    Key                         key;
    ACOM::Network::CallContext* call_cc; 
    ACOM::Network::Request*     req;      /* request to local post to clear if !=0 */
  };
  /**
  */
  Rendezvous* allocate_rendezvous();

  /**
  */
  void deallocate_rendezvous(Rendezvous*);
    
  /**
  */
  Rendezvous* get_rendezvous(uint32_t tag, uint32_t serial);

  /**
  */
  void erase_rendezvous( Rendezvous* rdv );

  /* Protected data member */
  LHashMap<Key, Rendezvous*> _allrdv; // list of rendez vous
};
 


/*
 * inline definitions
 */

}

#endif
