// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/NetworkCore/src/acom_netclient.h,v $
// Copyright(c)'2001 by Apache Team
// Authors: Jacques Briat, Thierry Gautier, Cyrille Martin
// ==========================================================================
#ifndef _ACOM_NETCLIENT_H_
#define _ACOM_NETCLIENT_H_
#include <acom_netbase.h>
#include <acom_iovect.h>
#include <acom_stream.h>

//@Include: acom_netconfig.h acom_comerror.h acom_types.h

// ACOM namespace
namespace ACOM_NET_NAMESPACE {


/* fwd declaration */
class Callback;

// ---------------------------------------------------------------------------
// NetworkClient class
// ---------------------------------------------------------------------------
/**@name NetworkClient
The virtual API for new Inuktitut interface Network.
The Network interface is splitted in serveral sub interfaces.
NetworkClient deals with sending message.
*/
class NetworkClient {
public:
  /** Default Cstor
  */
  NetworkClient();
  
  /** Destructor
  */
  virtual ~NetworkClient();
  
  // ---------------------------------------------------------------------------
  // Request class
  // --------------------------------------------------------------------------- 
  /**@name Request
    Format of request for the Network.  This class is the client side descriptor 
    of a send request. It registers the data to send and maintains the state of 
    progress of the sending actions.
    A request may be reused concurrently or sequentially.  It is garbaged when 
    the announced posts are all terminated. The callback is done before the 
    request deallocation.. 
  */
  class Request : public NetworkBase::BaseRequest {  
  public :  
    /** default constructor
    */
    Request();

    /** destructor
    */
    ~Request();

    /** initializer of request( allocator for iovector & buffers
    */
    bool init(  size_t bsz, ByteCellAllocator* bufa,
                int iovsz, IOVectCellAllocator* iova,
                Network* manager );

    /** finalizer of a request ; free all resources
    */
    void finish();
  
    /** signal a completed send of the request.
      Call the callback and release the request if no more used
    */
    void notify();

    /** pack immediate value ( return the size of packed data )
    */
    void packEos( );

    /** pack immediate value ( return the size of packed data )
    */
    size_t pack( const void* s, size_t ssz);

    /** pack an iovec entry in deferred sequence (true if OK)
    */
    bool put( IOVectEntry* s);

    /** pack  an iovec entry in deferred sequence (true if OK)
    */
    bool put( void * b, size_t sz);

    /** Accessor to internal data
    */
    ByteSeq* get_datastream();

    /** Accessor to internal data
    */
    IOVectSeq* get_iovstream();

    /** Return a pointer to the extended Network header
    */
    void* get_ext_header();
    
  public:
    /** default 1 when request is created.
        -1 when request iscompleted. 
        Then 0 callback is called
    */
    PTH::AtomicCounter _used_ct ; 

    /** Callback to call at local completion date */
    Callback* _f_cbk;   

    /** stream of immediate data */
    ByteSeq   _datastream;   

    /** stream of deferred data blocks */
    IOVectSeq _iovstream;   
  };




  // ---------------------------------------------------------------------------
  // Channel class
  // ---------------------------------------------------------------------------
  /**@name Channel
  The implementation of a specific OutChannel for a particular network
  should be done by making a derivate class of OutChannel for that network.
  The post method should be redefined in order to make a Network::Request.
  @see Network
  @see ProtocolService
  */
  class OutChannel {
  public:
    /** Default constructor */
    OutChannel();
    
    /** Destructor */
    virtual ~OutChannel();
        
    /** AM, AWS, WS protocols */
    virtual void post( Request* req ) throw (ComFailure)=0;
        
    /** Return the network's manager */
    Network* get_net();

    /** Return the peer node */
    NetworkBase::LocalNodeNo get_peer();
    
  protected: 
    /** Initialize the object */
    void init(Network* net, NetworkBase::LocalNodeNo peer);

  protected: 
    Network*                 _net;  // - medium for communication 
    NetworkBase::LocalNodeNo _peer; // - node with remote connection
    friend class Network;
  }; // -- end class Channel
  

  // ---------------------------------------------------------------------------
  // Network interface
  // ---------------------------------------------------------------------------

  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** Allocate a new request.
  */
  virtual Request* allocate_request ( ) throw (BadAlloc) =0;

  /** free request descriptor when unused
  */
  virtual void deallocate_request( Request* r ) throw (InvalidArgument) =0;

  /** Cancel a request allocated by allocate_request ( true == cancel done)
      Cancelation of request depend on state of progress
      @param req (IN)
  */
  virtual bool cancel_request( Request* req ) 
    throw (InvalidArgument);

  /** send  to a node (AM,AWS,WS protocols) 
      @param   req  request (IN) 
       @param   ch  channel  (IN) 
  */
  virtual void post( Request* req, OutChannel* ch=0 ) 
       throw (ComFailure);

protected:
  friend class BaseRequest;
  friend class OutChannel;
  friend class Request;
  friend class Receipt;
}; // -- end class Network


/*
 * inline definition
 *
 */
// ---------------------------------------------------------------------------
// Request class
// --------------------------------------------------------------------------- 
inline NetworkClient::Request::Request()
  : BaseRequest(), _used_ct(0),
    _f_cbk(0), _datastream(), _iovstream() 
{}

inline NetworkClient::Request::~Request() 
{}

inline bool NetworkClient::Request::init(
  size_t bsz, ByteCellAllocator* bufa,
  int iovsz, IOVectCellAllocator* iova,
  Network* manager 
)
{ 
  BaseRequest::init(manager);
  if (!_iovstream.init( iovsz, iova )) 
    return false;
  if (!_datastream.init( bsz, bufa, iovsz, iova))
    { _iovstream.destroy(); return false; }
  return true;
}

inline void NetworkClient::Request::finish() 
{ 
  _state = NetworkBase::S_DESTROY; _error_no = ComFailure::OK;
  _datastream.destroy(); _iovstream.destroy(); 
}

inline void NetworkClient::Request::packEos( )
{ _datastream.packEos(); }

inline size_t NetworkClient::Request::pack( const void* s, size_t ssz)
{ return _datastream.pack( s, ssz ); }

inline bool NetworkClient::Request::put( IOVectEntry* s) 
{ return _iovstream.put(s); }

inline bool NetworkClient::Request::put( void * b, size_t sz) 
{ return _iovstream.put(b,sz); }

inline ByteSeq* NetworkClient::Request::get_datastream() 
{ return &_datastream; }

inline IOVectSeq* NetworkClient::Request::get_iovstream() 
{ return &_iovstream; }

inline void* NetworkClient::Request::get_ext_header()
{ return _ext_header; }


// ---------------------------------------------------------------------------
// Channel class
// ---------------------------------------------------------------------------
inline NetworkClient::OutChannel::OutChannel() : _net(0) 
{}

inline NetworkClient::OutChannel::~OutChannel() 
{}
    
inline Network* NetworkClient::OutChannel::get_net() 
{ return _net; }

inline NetworkBase::LocalNodeNo NetworkClient::OutChannel::get_peer() 
{ return _peer; }

inline void NetworkClient::OutChannel::init(Network* net, NetworkBase::LocalNodeNo peer)
{ _net=net; _peer=peer; }



} // end namespace ATH...
#endif 
