//
// $Id: mput0.cpp,v 1.2 2003/07/28 20:31:18 thierry Exp $
// Author : T Gautier,  INRIA
// Illustration of CallContext on a parallel mput command that
// duplicate a file on different machine with different file systems.
// Basic algorithm for broadcast is a chain
// * input : input filename, output filename
// See testcc2.cpp for simple use of CallContext

#include <acom>
#include <pth>
#include <list>
#include <cstdio>
#include <iostream>
#include <iomanip>

#include <sys/types.h>  // mmap, stat
#include <sys/mman.h>   // mmap
#include <fcntl.h>      // open 
#include <unistd.h>     // lseek, close
#include <sys/stat.h>   // stat


/* Size of the packet: 
*/
static int SIZE_PACKET  = 8192;


/* The network used for all communications
*/
static ACOM::Network* net = 0;
static int rank = 0;
static int nodecount = 0;

/* ===========================================================================
 * Definition the services and thread's class for redirection
 * =========================================================================== */

/* Ack service that send ack to node 0
*/
class AckService: public ACOM::ServiceCC {
public:
  /* Cstor */
  AckService();
  
  /* Dstor */
  ~AckService();

  /* wait ack */
  void wait();

protected:
  PTH::Condition _cond;
  PTH::Mutex     _access;
  bool           _ok;
  
  /* This is low level entry point of a service : 
  */
  void server_call( ACOM::Network::CallContext* cc ) 
    throw(ACOM::ComFailure);
};
static AckService* ack_service =0;


/* Echo service : recv message and push it (the call context cc)
   to the thread for redirection. This service is invoked using
   the ActiveMessage protocol (for the first message : an header)
   and using the WriteAndSignal (for the all other messages).
*/
class EchoService: public ACOM::ServiceCC {
public:
  /* Cstor */
  EchoService();
  
  /* Dstor */
  ~EchoService();

protected:
  void*      _mmap_data;  /* pointer to the begin of the memory map output file */
  PTH::Mutex _access;

  /* This is low level entry point of a service : 
  */
  void server_call( ACOM::Network::CallContext* cc ) 
    throw(ACOM::ComFailure);

  /* This is low level entry point of a service : 
  */
  void server_get_iovect( ACOM::Network::CallContext* cc ) 
    throw(ACOM::ComFailure);
    
  friend class ThreadRedirec;
};
static EchoService* echo_service =0;


/* Callback called when all requests are sent
*/
class Mycallback : public ACOM::Callback {
public:
  /* */
  Mycallback(PTH::Mutex* l, PTH::Condition* c, ACOM::Network::CallContext* cc) 
   : _lock(l), _cond(c), _cc(cc), _is_called(false) 
  {}

  /* Called when a request could be delete and data 
     could be reuse.
  */
  void done( ACOM::Network::Request* );
  
  /* Wait the callback was called
  */
  void wait();


protected:
  PTH::Mutex*                 _lock;
  PTH::Condition*             _cond;
  ACOM::Network::CallContext* _cc;
  bool                        _is_called;
};


/* thread that handles list of CallContext to redirect them to next node
*/
class ThreadRedirec {
public:
  /* build a thread that will redirect a message to dest nn */
  ThreadRedirec( );
  
  /* destroy the thread objet */
  ~ThreadRedirec( );
  
  /* Ping service push a new cc to redirect, that will
     be re-send by the current thread.
     Push in back, pop in front to conserve fifo
  */
  void push( ACOM::Network::CallContext* cc );
  
  /* Consumer of cc, entry point of the thread */
  void run();
  
  /* Stop the thread */
  void stop();
  
public:
  /* entry point for PTH thread creation */
  static void* run(void* arg);

  class CommonHeader {
  public:
    size_t serial;
    off_t  offset;
    std::ostream& print( std::ostream& o )
    { return o << "serial=" << serial << ", offset=" << offset; }
  };
  class FirstHeader : public CommonHeader {
  public:
    mode_t mode;
    int    sz_filename1;
    char*  filename1;
    int    sz_filename2;
    char*  filename2;
    std::ostream& print( std::ostream& o )
    { return CommonHeader::print(o) << ", mode=" << mode 
                                    << ", sz_filename1=" << sz_filename1
                                    << ", filename1=" << filename1
                                    << ", sz_filename2=" << sz_filename2
                                    << ", filename2=" << filename2;
    }
  };
  class PacketHeader : public CommonHeader {
  public:
    size_t size_packet;
    off_t  size_remain;
    std::ostream& print( std::ostream& o )
    { return CommonHeader::print(o) << ", size_packet=" << size_packet 
                                    << ", size_remain=" << size_remain;
    }
  };
  
protected:
  typedef ACOM::Network::CallContext* OneCC;

  PTH::Condition   _cond;
  PTH::Condition   _condstop;
  PTH::Mutex       _access;  
  bool             _wait;         /* true if the thread wait a cc */
  bool             _stop;         /* true if the thread should stop */
  std::list<OneCC> _cc_toredirec; /* list of cc to redirect */  
};
static ThreadRedirec* thread =0;



/* ===========================================================================
 * Implantation of previous methods
 * =========================================================================== */

/*
*/
AckService::AckService()
{
  _cond.init();
  _access.init();
  _ok = false;
}
  

/*
*/
AckService::~AckService()
{
  _cond.destroy();
  _access.destroy();
}


/*
*/
void AckService::wait()
{
  PTH::CriticalGuard critical( _access );
  while (!_ok)
    _cond.wait( _access ); 
  _ok = false;
}


/*
*/
void AckService::server_call( ACOM::Network::CallContext* cc ) 
  throw(ACOM::ComFailure)
{
  PTH::CriticalGuard critical( _access );
  _ok = true;
  _cond.signal();
}


/*
*/
EchoService::EchoService()
 : _mmap_data(0)
{
  _access.init();
}
  

/*
*/
EchoService::~EchoService()
{
  _access.destroy();
}


/*
*/
void EchoService::server_call( ACOM::Network::CallContext* cc ) 
  throw(ACOM::ComFailure)
{
  if (cc->error_no != ACOM::ComFailure::OK) 
    std::cerr << "Error in msg='" << cc->error_no << std::endl;

  if (cc->cc ==0) {
    /* first packet : lock the service util the 
       output filename was created
    */
    _access.lock();
  }  
  thread->push( cc );
}


/*
*/
void EchoService::server_get_iovect( ACOM::Network::CallContext* cc ) 
  throw(ACOM::ComFailure)
{
  /* file iov[0] by the address of the offset of the packet */
  ThreadRedirec::PacketHeader* h = 
      (ThreadRedirec::PacketHeader*)cc->push(sizeof(ThreadRedirec::PacketHeader) );
  cc->cc = h;
  cc->receipt.unpack( &h->serial, sizeof(size_t) );
  cc->receipt.unpack( &h->offset, sizeof(off_t) );
  cc->receipt.unpack( &h->size_packet, sizeof(size_t) );
  cc->receipt.unpack( &h->size_remain, sizeof(off_t) );

  PTH::CriticalGuard critical(_access);
  cc->iov[0].data = h->offset + (char*)_mmap_data;
}


/*
*/
void Mycallback::done( ACOM::Network::Request* )
{
  { PTH::CriticalGuard critical(*_lock);
    _is_called = true;
  }
  _cond->signal();
  _cc->release();
}


/* 
*/
void Mycallback::wait()
{
  PTH::CriticalGuard critical(*_lock);
  while (!_is_called)
    _cond->wait(*_lock);
}


/* build a thread that will redirect a message to dest nn
*/
ThreadRedirec::ThreadRedirec( )
{
  _access.init();
  _cond.init();
  _condstop.init();
  _wait = false;
  _stop = false;
  _cc_toredirec.clear();
}


/*
*/
ThreadRedirec::~ThreadRedirec( )
{
  _access.destroy();
  _cond.destroy();
  _condstop.destroy();
  _cc_toredirec.clear();
}


/*
*/
void ThreadRedirec::stop()
{
  PTH::CriticalGuard critical(_access);
  while (!_stop) {
    _condstop.wait(_access);
  }
}


/*
*/
void ThreadRedirec::push( ACOM::Network::CallContext* cc )
{
  PTH::CriticalGuard critical(_access);
  _cc_toredirec.push_back( cc );
  cc->acquire();
  if (_wait) _cond.signal();
}

  
/* 
*/
void ThreadRedirec::run()
{
  /* redirect the cc to nextnode */
  int nextnode = (rank + 1) % nodecount;
  ACOM::Network::OutChannel* channel = net->get_default_route( nextnode );
  ACOM::Network::CallContext* cc = 0;
  
  int count = 0;
  int out_fd;
  size_t count_packet_recv = 0;     /* count of received packets */
  size_t count_packet_toberecv = 0; /* count of packets to be received */
  off_t size_file = 0;

  /* this is a daemon's like thread, infinite loop except if it stop */
  bool cestfini = false;
  while (!cestfini)
  {
    /* get a cc to redirect */
    { PTH::CriticalGuard critical(_access);
      if (_cc_toredirec.empty()) 
      {
        _wait = true;
        while (_cc_toredirec.empty()) 
          _cond.wait( _access );
        _wait = false;
      }
      cc = _cc_toredirec.front();
      _cc_toredirec.pop_front();
    }
    
    if (cc->cc ==0) 
    { /* first packet : create the file and memory map it */
      FirstHeader h;
      cc->receipt.unpack( &h.serial, sizeof(size_t) );
      cc->receipt.unpack( &h.offset, sizeof(off_t) );
      cc->receipt.unpack( &h.mode, sizeof(mode_t) );
      cc->receipt.unpack( &h.sz_filename1, sizeof(int) );
      h.filename1 = (char*)cc->allocate( h.sz_filename1 );
      cc->receipt.unpack( h.filename1, h.sz_filename1 );
      cc->receipt.unpack( &h.sz_filename2, sizeof(int) );
      h.filename2 = (char*)cc->allocate( h.sz_filename2 );
      cc->receipt.unpack( h.filename2, h.sz_filename2 );
      //h.print(std::cout << rank << ":receive first header=") << std::endl;
      
      /* compute the number of packet to receive */
      size_file = h.offset;
      count_packet_toberecv = (size_file + SIZE_PACKET -1) / SIZE_PACKET;
      
      /* resend it if nextnode !=0 */
      if (nextnode !=0) {
        /* callback to wait posted message, call back make a call to release */      
        cc->acquire();
        Mycallback mycbk( &_access, &_cond, cc);

        ACOM::ActiveMessage<EchoService> am (net, echo_service, &mycbk);
        am->pack( &h.serial, sizeof(size_t) );
        am->pack( &size_file, sizeof(off_t) );
        am->pack( &h.mode, sizeof(mode_t) );
        am->pack( &h.sz_filename1, sizeof(int) );
        am->pack( h.filename1, h.sz_filename1 );
        am->pack( &h.sz_filename2, sizeof(int) );
        am->pack( h.filename2, h.sz_filename2 );
        am.post(channel);
        //std::cout << "Resent file1=" << h.filename1 << ", file2=" << h.filename2 << std::endl;
        /* wait message is really sent */
        mycbk.wait();
      }

      /* open local file: add the rank as suffix in the name */
      char filename[256];
      sprintf(filename, "%s.%i", h.filename2, rank);
      out_fd = open(filename, O_CREAT|O_TRUNC|O_RDWR, h.mode);
      if (out_fd <0) {
        echo_service->_access.unlock();
        ACOM::throwit( ACOM::Runtime("cannot create file", errno) );
      }
      fchmod( out_fd, h.mode);
      std::cout << "File: " << filename << " created, begin copie" << std::endl;
      
      /* fix the size of the filename */
      lseek(out_fd, size_file-1, SEEK_SET);
      write( out_fd, filename, 1 );

      /* memory map it */
      echo_service->_mmap_data = mmap( 0, size_file, PROT_WRITE|PROT_READ, MAP_SHARED, out_fd, 0);
      if (echo_service->_mmap_data == (void*)-1) {
        echo_service->_access.unlock();
        ACOM::throwit( ACOM::Runtime("cannot memory map the file", errno) );
      }
/*
      std::cout << "File data mapped at @=" << echo_service->_mmap_data << std::endl;
      for ( char* addr = (char*)echo_service->_mmap_data; 
            addr < size_file + (char*)echo_service->_mmap_data;
            addr += 4096 )
        *addr = 0;
*/
      
      /* ok file is create and memory mapped, wait recev of data, unlock service */
      echo_service->_access.unlock();
      
      /* test if end ? */
      cestfini = (h.offset ==0);
    }
    else {
      /* data packet */
      PacketHeader* h = (PacketHeader*)cc->cc;
      ++count_packet_recv;
      /* resend it if nextnode !=0 */
      if (nextnode !=0) {
        ACOM::WriteAndSignal<EchoService> msg (net, echo_service, 0);
        msg->pack( &h->serial, sizeof(size_t) );
        msg->pack( &h->offset, sizeof(off_t) );
        msg->pack( &h->size_packet, sizeof(size_t) );
        msg->pack( &h->size_remain, sizeof(off_t) );
        msg->put( &cc->iov[0] );
        /* post the message */
        msg.post( channel );
      }
      
      /* do not need to write it. Due to the memory mapping of the output 
         file, the Inuktitut network extract correct packet at the correct address,
         even if fifo order is not respected !
      */
      /* WARN : here assumption is done about the fact that the message
         is send before we release the call context
      */
      /* test if end ? if FIFO order of insertion into list of thread
         then the test could be replaced by h->size_remain ==0
      */
      cestfini = (count_packet_recv == count_packet_toberecv);
    }
    
    /* release the cc */
    cc->release();
  }

final_ack:
  /* free memory */
  //msync( echo_service->_mmap_data, size_file, MS_SYNC );
  if (munmap( echo_service->_mmap_data, size_file) <0)
    ACOM::throwit( ACOM::Runtime("cannot unmap file in memory ", errno) );
  close(out_fd);

  /* final node : send ack to */
  if (nextnode ==0) {
    ACOM::Network::OutChannel* channel = net->get_default_route(0);
    ACOM::ActiveMessage<AckService> am (net, ack_service, 0);
    am.post(channel);
  }
  std::cout << rank << "::End copy" << std::endl;
}


/* 
*/
void* ThreadRedirec::run(void* arg) 
{
  ThreadRedirec* obj = (ThreadRedirec*)arg;
  try {
    obj->run();
    PTH::CriticalGuard critical(obj->_access);
    obj->_stop = true;
    obj->_condstop.signal();
  } catch (const ACOM::Runtime& ex) {
    std::cout << "Catch error= " << ex.what() << std::endl;
  }
  return 0;
}
  


/* ===========================================================================
 * Do it method
 * =========================================================================== */
void doit ( const std::string& in_filename, const std::string& out_filename )
{
  if (out_filename == "")
    ACOM::throwit( ACOM::Runtime("bad output filename") );

  /* Open file and get its size
  */
  int in_fd = open(in_filename.c_str(), O_RDONLY, 0);
  if (in_fd <0) 
    ACOM::throwit( ACOM::Runtime("cannot open file", errno) );
  struct stat sb;
  fstat(in_fd, &sb );
  
  /* compute size of the file 
  */
  off_t size_file = sb.st_size;
  
  /* memory map the file 
  */
  void* mmap_data = mmap( 0, size_file, PROT_READ, MAP_SHARED|MAP_FILE, in_fd, 0);
  char* data = (char*) mmap_data;
  if (data == (char*)-1)
    ACOM::throwit( ACOM::Runtime("cannot map file in memory ", errno) );

  /* Get a channel, form the message
  */
  ACOM::Network::OutChannel* channel = net->get_default_route(1);

  /* Set first packet : ActiveMessage
     - IA: serial number
     - IA: size of the file
     - IA: 1+size of in_filename as C string
     - IA: 1+size char
     - IA: 1+size of out_filename as C string
     - IA: 1+size char
  */
  size_t serial =0;
  { 
    ACOM::ActiveMessage<EchoService> am (net, echo_service, 0);
    am->pack( &serial, sizeof(size_t) );
    am->pack( &size_file, sizeof(off_t) );
    am->pack( &sb.st_mode, sizeof(mode_t) );
    /* input filename */
    int sz = strlen(in_filename.c_str())+1;
    am->pack( &sz, sizeof(int) );
    am->pack( in_filename.c_str(), sz );
    /* output filename */
    sz = strlen(out_filename.c_str())+1;
    am->pack( &sz, sizeof(int) );
    am->pack( out_filename.c_str(), sz );
    am.post(channel);
  }

  /* Send packets : Write and signal
     - IA: serial number
     - IA: size of the packet
     - IA: offset of the packet
     - IA: size remains to send (0 -> last packet)
     - DA: the packet
  */
  off_t size = size_file;
  off_t off_packet = 0;
  while (size >0) {
    ++serial;
    size_t size_packet = ( size > SIZE_PACKET ? SIZE_PACKET : size );
    size -= size_packet;
    off_packet = (char*)data - (char*)mmap_data;
    ACOM::WriteAndSignal<EchoService> msg (net, echo_service, 0);
    msg->pack( &serial, sizeof(size_t) );
    msg->pack( &off_packet, sizeof(off_t) );
    msg->pack( &size_packet, sizeof(size_t) );
    msg->pack( &size, sizeof(off_t) );
    msg->put( data, size_packet );
    data += size_packet;
    msg.post(channel);
  }

  /* wait end of operation */
  ack_service->wait();
  
  /* free memory */
  if (munmap( mmap_data, size_file) <0)
    ACOM::throwit( ACOM::Runtime("cannot unmap file in memory ", errno) );
  close(in_fd);
}


/* ===========================================================================
 * Main
 * =========================================================================== */
int main (int argc, char* argv[])
{
  try {
    /* Initialize network */
    ACOM::Acom::initialize( &argc, &argv );
    
    /* create the service objects */
    ack_service = new AckService;
    echo_service = new EchoService;
    ACOM::Service::Manager.bind (2, ack_service);
    ACOM::Service::Manager.bind (3, echo_service);

    /* Commit network */
    ACOM::Acom::commit( );

    /* Get the network:
    */
    net = ACOM::Acom::get_default_network();

    /* BUG : cannot have the correct node number with SOCKNET if no barrier here ? */
    net->barrier();
    
    /* rank and nodecount */
    rank = net->get_rank();
    nodecount = net->get_nodecount();

    /* create the redirection thread */
    if (rank !=0) {
      thread = new ThreadRedirec;
      PTH::ThreadId tid;
      PTH::Thread::create( &tid, &ThreadRedirec::run, (void*)thread);
    }
    
    /* Wait every body start its thread */
    net->barrier();

      
    if (nodecount >1) {
      /* Send : COUNT_PACKET of size SIZE_PACKET */
      if (rank ==0) {
        /* Get the argument
        */
        std::string in_filename  = argv[1];
        std::string out_filename = argv[2];
        doit(in_filename, out_filename );
        std::cout << "Copy: " << in_filename << " into " << out_filename << std::endl;
      }
      if (thread !=0)
        thread->stop();
      net->barrier();
    }

    /* Terminate */
    ACOM::Acom::terminate( );

    /* Delete services */
    delete echo_service;
    delete ack_service;

  } catch (ACOM::Runtime& ex) {
    std::cerr << "Catch exception: " << ex.what() << std::endl;
  }
  catch (...) {
    std::cerr << "Unknown exception" << std::endl;
  }
  return 0;
}
