//
// $Id: mput1.cpp,v 1.6 2003/07/29 20:01:55 thierry Exp $
// Author : T Gautier,  INRIA
// Illustration of ICS broadcast capabilities on the
// duplicate a file on different machines.
// Arguments are <input file> <output file> currently the output
// file is number by the rank of the Inuktitut's node.
// The method is the following : first the node of rank 0 (those
// which have the input file> send a message (active message)
// to the EchoService to create the file on each remote node.
// Then it wait until all other nodes have created the file
// and memory mapped it.
// The file is send by packet of size specified by the option
// -b <packet size> on the command line. By default this size
// is 2048 bytes. Each packet is sent with the offset from the
// begining of the file. The used protocol is write and signal.
// The service (the EchoService) decode the address by returning
// the base address of the file in memory + the offset of the
// packet. Using this method and what ever is the order
// of message (especially with ICS that can used multiple threads
// to process output and input message), the file is filling
// correctly. The last message close the file and un map it. 
// The broadcast algorithm is specified as -m argument.
// Run executable with -h to display options.
// * input : input filename, output filename
// See mput0.cpp for same example at low level of interface.

#include <acom>
#include <pth>
#include <ics>
#include <cstdio>
#include <iostream>
#include <iomanip>

#include <sys/types.h>  // mmap, stat
#include <sys/mman.h>   // mmap
#include <fcntl.h>      // open 
#include <unistd.h>     // lseek, close
#include <sys/stat.h>   // stat


/* The network used for all communications
*/
static ACOM::Network*     net = 0;         /* default network */
static int                rank = 0;
static int                nodecount = 0;
static ICS::BcastNetwork* bcastnet = 0;    /* broadcast network on top of net */
static int                tag_channel = 3; /* id of the channel for all nodes */
static int                key_ack     = 8; /* key used for ack service */
static ICS::AckService*   ack_service;     /* service to send back an ack */
static ICS::DefaultBarrier* barrier_service;  /* service for fast barrier */

/* command line options 
*/
static std::string        methodname = "flat"; /* the name of the method, flat, chain, alpha */
static double             alpha = 0.5;         /* alpha value */
static size_t             packetsize = 2048;   /* size of bloc for reading the file */
static bool               verboseon = false;   /* verbose mode */

/* fwd declarations */
void Usage(const std::string& msg ="");
int Bcast_parseopt( int argc, char* argv[] );


/* ===========================================================================
 * Definition the services and thread's class for redirection
 * =========================================================================== */


/* Echo service : recv message and push it (the call context cc)
   to the thread for redirection. This service is invoked using
   the ActiveMessage protocol (for the first message : an header)
   and using the WriteAndSignal (for the all other messages).
*/
class EchoService: public ACOM::ServiceCC {
public:
  /* Cstor */
  EchoService();
  
  /* Dstor */
  ~EchoService();

  class CommonHeader {
  public:
    size_t serial;
    off_t  offset;
    std::ostream& print( std::ostream& o )
    { return o << "serial=" << serial << ", offset=" << offset; }
  };
  class FirstHeader : public CommonHeader {
  public:
    mode_t mode;
    int    sz_filename1;
    char*  filename1;
    int    sz_filename2;
    char*  filename2;
    std::ostream& print( std::ostream& o )
    { return CommonHeader::print(o) << ", mode=" << mode 
                                    << ", sz_filename1=" << sz_filename1
                                    << ", filename1=" << filename1
                                    << ", sz_filename2=" << sz_filename2
                                    << ", filename2=" << filename2;
    }
  };
  class PacketHeader : public CommonHeader {
  public:
    size_t size_packet;
    off_t  size_remain;
    std::ostream& print( std::ostream& o )
    { return CommonHeader::print(o) << ", size_packet=" << size_packet 
                                    << ", size_remain=" << size_remain;
    }
  };

protected:
  void*      _mmap_data;     /* pointer to the begin of the memory map output file */
  int        _out_fd;        /* local file descriptor of the output file */
  off_t      _size_file;

  /* This is low level entry point of a service : 
  */
  void server_call( ACOM::Network::CallContext* cc ) 
    throw(ACOM::ComFailure);

  /* This is low level entry point of a service : 
  */
  void server_get_iovect( ACOM::Network::CallContext* cc ) 
    throw(ACOM::ComFailure);
};
static EchoService* echo_service =0;



/* ===========================================================================
 * Implantation of previous methods
 * =========================================================================== */

/*
*/
EchoService::EchoService()
 : _mmap_data(0), _size_file(0)
{
}
  

/*
*/
EchoService::~EchoService()
{
}


/*
*/
void EchoService::server_call( ACOM::Network::CallContext* call_cc ) 
  throw(ACOM::ComFailure)
{
  if (verboseon)
    std::cout << "IN: " << __PRETTY_FUNCTION__ << std::endl;
  if (call_cc->cc ==0) 
  { 
    /* first packet : create the file and memory map it */
    FirstHeader h;
    call_cc->receipt.unpack( &h.serial, sizeof(size_t) );
    call_cc->receipt.unpack( &h.offset, sizeof(off_t) );
    call_cc->receipt.unpack( &h.mode, sizeof(mode_t) );
    call_cc->receipt.unpack( &h.sz_filename1, sizeof(int) );
    h.filename1 = (char*)call_cc->allocate( h.sz_filename1 );
    call_cc->receipt.unpack( h.filename1, h.sz_filename1 );
    call_cc->receipt.unpack( &h.sz_filename2, sizeof(int) );
    h.filename2 = (char*)call_cc->allocate( h.sz_filename2 );
    call_cc->receipt.unpack( h.filename2, h.sz_filename2 );
    if (verboseon)
      h.print(std::cout << rank << ":receive first header=") << std::endl;
    
    /* compute the number of packet to receive */
    _size_file = h.offset;
    
    /* open local file: add the rank as suffix in the name */
    char filename[256];
    sprintf(filename, "%s.%i", h.filename2, rank);
    _out_fd = open(filename, O_CREAT|O_TRUNC|O_RDWR, h.mode);
    if (_out_fd <0) {
      ACOM::throwit( ACOM::Runtime("cannot create file", errno) );
    }
    fchmod( _out_fd, h.mode);
    std::cout << "File: " << filename << " created, begin copie" << std::endl;
    
    /* fix the size of the filename */
    lseek(_out_fd, _size_file-1, SEEK_SET);
    write( _out_fd, filename, 1 );

    /* memory map it */
    _mmap_data = mmap( 0, _size_file, PROT_WRITE|PROT_READ, MAP_SHARED, _out_fd, 0);
    if (_mmap_data == (void*)-1) {
      ACOM::throwit( ACOM::Runtime("cannot memory map the file", errno) );
    }    
    ack_service->signal(0, key_ack+1 );

    if (verboseon)
      std::cout << "OUT: " << __PRETTY_FUNCTION__ << std::endl;
    /* ok return from the service invocation */
    return;
  }

  /* it is the last packet ? */
  PacketHeader* h = (PacketHeader*)call_cc->cc;
  if (verboseon) {
    char* data = (char*)call_cc->iov[0].data;
    std::cout << "\n-------- begin packet: " << h->serial << std::endl;
    for (int i=0; i< call_cc->iov[0].len; ++i)
      std::cout << data[i];
    std::cout << "\n-------- end packet: " << h->serial << std::endl;
  }
  
  if (h->size_remain >0) {
    if (verboseon) 
      std::cout << "OUT: " << __PRETTY_FUNCTION__ << std::endl;
    return;
  }

  /* munmap, close and signal node 0 */
  if (munmap( _mmap_data, _size_file) <0)
    ACOM::throwit( ACOM::Runtime("cannot unmap file in memory ", errno) );
  close(_out_fd);

  /* send to 0 the ack */
  ack_service->signal(0, key_ack );
  std::cout << rank << "::End copy" << std::endl;
  if (verboseon) 
    std::cout << "OUT: " << __PRETTY_FUNCTION__ << std::endl;
}


/* Each bloc of the file is send using write and signal:
   the method decodes address using the offset of the
   packet in the original file, the returned address (iovector)
   is the address where to write the memory mapped file.
   Nothing else is necessary to write the file. 
   Even if two packets are not receive in the same order they have
   be sent, the packets are write to their correct location.
*/
void EchoService::server_get_iovect( ACOM::Network::CallContext* call_cc ) 
  throw(ACOM::ComFailure)
{
  if (verboseon) 
    std::cout << "IN: " << __PRETTY_FUNCTION__ << std::endl;
  /* file iov[0] by the address of the offset of the packet */
  PacketHeader* h = (PacketHeader*)call_cc->allocate(sizeof(PacketHeader) );
  call_cc->cc = h;
  call_cc->receipt.unpack( &h->serial, sizeof(size_t) );
  call_cc->receipt.unpack( &h->offset, sizeof(off_t) );
  call_cc->receipt.unpack( &h->size_packet, sizeof(size_t) );
  call_cc->receipt.unpack( &h->size_remain, sizeof(off_t) );

  if (verboseon) 
    h->print(std::cout << "PackerHeader received:" ) << std::endl;
  call_cc->iov[0].data = h->offset + (char*)_mmap_data;
//  std::cout << "return iov @: " << call_cc->iov[0].data << std::endl;
  if (verboseon) 
    std::cout << "OUT: " << __PRETTY_FUNCTION__ << std::endl;
}



/* ===========================================================================
 * Do it method
 * =========================================================================== */
void doit ( const std::string& in_filename, const std::string& out_filename )
{
  if (out_filename == "")
    ACOM::throwit( ACOM::Runtime("bad output filename") );

  /* Open file and get its size
  */
  int in_fd = open(in_filename.c_str(), O_RDONLY, 0);
  if (in_fd <0) 
    ACOM::throwit( ACOM::Runtime("cannot open file", errno) );
  struct stat sb;
  fstat(in_fd, &sb );
  
  /* compute size of the file 
  */
  off_t size_file = sb.st_size;
  
  /* memory map the file 
  */
  void* mmap_data = mmap( 0, size_file, PROT_READ, MAP_SHARED|MAP_FILE, in_fd, 0);
  char* data = (char*) mmap_data;
  if (data == (char*)-1)
    ACOM::throwit( ACOM::Runtime("cannot map file in memory ", errno) );

  /* Get the broadcast channel
  */
  ICS::Network::OutChannel* bcastchannel = bcastnet->get_channel(tag_channel);

  /* Set first packet : ActiveMessage
     - IA: serial number
     - IA: size of the file
     - IA: 1+size of in_filename as C string
     - IA: 1+size char
     - IA: 1+size of out_filename as C string
     - IA: 1+size char
  */
  size_t serial =0;
  //ACOM::ActiveMessage<EchoService> am (bcastnet, echo_service, 0);
  ACOM::Message am (bcastnet, ACOM::Network::P_AM, echo_service, 0);
  am->pack( &serial, sizeof(size_t) );
  am->pack( &size_file, sizeof(off_t) );
  am->pack( &sb.st_mode, sizeof(mode_t) );
  /* input filename */
  int sz = strlen(in_filename.c_str())+1;
  am->pack( &sz, sizeof(int) );
  am->pack( in_filename.c_str(), sz );
  /* output filename */
  sz = strlen(out_filename.c_str())+1;
  am->pack( &sz, sizeof(int) );
  am->pack( out_filename.c_str(), sz );
  
  /* broadcast it */
  am.post(bcastchannel);

  /* wait header is ok and file created */
  ack_service->wait(key_ack+1, nodecount-1); 
  
  if (verboseon) 
    std::cout << "Header sended" << std::endl;

  /* Send packets : Write and signal
     - IA: serial number
     - IA: size of the packet
     - IA: offset of the packet
     - IA: size remains to send (0 -> last packet)
     - DA: the packet
  */
  off_t size = size_file;
  off_t off_packet = 0;
  while (size >0) {
    ++serial;
    size_t size_ofthispacket = ( size > packetsize ? packetsize : size );
    size -= size_ofthispacket;
    off_packet = (char*)data - (char*)mmap_data;
//    ACOM::WriteAndSignal<EchoService> msg (bcastnet, echo_service, 0);
    ACOM::Message msg (bcastnet, ACOM::Network::P_WS, echo_service, 0);
    msg->pack( &serial, sizeof(size_t) );
    msg->pack( &off_packet, sizeof(off_t) );
    msg->pack( &size_ofthispacket, sizeof(size_t) );
    msg->pack( &size, sizeof(off_t) );
    msg->put( data, size_ofthispacket );
    data += size_ofthispacket;

    /* broadcast it */
    msg.post(bcastchannel);
    if (verboseon) 
      std::cout << "Packet " << serial << " sended" << std::endl;
  }

  /* wait end of operations : all node on the last packet send back an ack to 0 */
  ack_service->wait(key_ack, nodecount -1);
  
  /* free memory */
  if (munmap( mmap_data, size_file) <0)
    ACOM::throwit( ACOM::Runtime("cannot unmap file in memory ", errno) );
  close(in_fd);
}


/* ===========================================================================
 * Main
 * =========================================================================== */
int main (int argc, char* argv[])
{
  try {
    /* Initialize network */
    ACOM::Acom::initialize( &argc, &argv );

    int begarg = Bcast_parseopt( argc, argv );
    
    
    /* create the service objects */
    ack_service = new ICS::AckService;
    barrier_service = new ICS::DefaultBarrier;
    echo_service = new EchoService;
    
    ACOM::Service::Manager.bind (1, ack_service);
    ACOM::Service::Manager.bind (2, barrier_service);
    ACOM::Service::Manager.bind (3, echo_service);

    /* Commit network */
    ACOM::Acom::commit( );


    if (true) { //verboseon) {
      std::cout << "[Acom]:: initialize ok, mypid=" << ACOM::Acom::get_pid() << std::endl;
    }


    //short c; std::cin >> c;

    if (verboseon)
      std::cout << "[Acom]:: commit ok" << std::endl;

    /* Get the network  */
    net = ACOM::Acom::get_default_network( );
    if (net ==0) 
      ACOM::throwit( ACOM::InvalidArgument("undefined default network") );
    else if (verboseon)
      std::cout << "Use network name=" << net->get_name() << std::endl;

//    net->barrier();
    
    /* initialize ack service */
    ack_service->initialize(net);
    
    /* initialize barrier service */
    barrier_service->initialize(net);
    
    /* Get rank and nodecount */
    rank      =  net->rank();
    nodecount =  net->node_count();
    
    if (verboseon) {
      std::cout << "#node=" << nodecount << std::endl;
      std::cout << "rank =" << rank << std::endl;
    }
    
    
    /* if #nodes <= 1 do nothing */
    if (nodecount <=1) {
      ACOM::Acom::terminate();
      return 0;
    }

    /* build the network with broadcast depending of the algorithm  
       choose : flat tree, chain, alpha tree (binomial if alpha ==0.5)
    */
    if ( methodname == "flat" ) {
      bcastnet = new ICS::BcastFlatNetwork;
    } else if ( methodname == "chain" ) {
      bcastnet = new ICS::BcastChainNetwork;
    } else if ( methodname == "alpha" ) {
      bcastnet = new ICS::BcastAlphaNetwork;
    } else {
      Usage("Bad method");
    }
    if (rank ==0)
    {//if (verboseon) {
      std::cout << "Use network=" << net->get_name() 
                << ", algorithm=" << methodname 
                << ", packetsize=" << packetsize 
                << std::endl;
    }

    /* initialize the broadcast network */
    bcastnet->initialize(&argc, &argv, net);

    /* Build a broadcast channel that contains all nodes.
       The name of the channel is tag_channel.
    */
    ICS::uint16_t ranks[ nodecount ];
    for (int i=0; i<nodecount; ++i)
      ranks[i] =i;

    bcastnet->open_channel(tag_channel, nodecount, ranks, rank );
    if ( methodname == "alpha") {
      ICS::BcastAlphaNetwork::OutChannel* ch = (ICS::BcastAlphaNetwork::OutChannel*)bcastnet->get_channel( tag_channel );
      ch->set_alpha( alpha );
    }

    /* Here broadcast could begin : we should ensure that all nodes
       have declared the broadcast channel, ie the name is well
       known on all nodes
    */
    barrier_service->barrier(10);

    if (rank ==0) {
      /* Get the argument
      */
      std::string in_filename  = argv[begarg];
      if (verboseon)
        std::cout << "Input filename:" << in_filename << std::endl;
      std::string out_filename = argv[begarg+1];
      if (verboseon)
        std::cout << "Output filename:" << out_filename << std::endl;
    
      std::cout << "Copy: " << in_filename << " into " << out_filename << std::endl;
      /* Send : COUNT_PACKET of size packetsize */
      doit(in_filename, out_filename );
    }
    
    /* Wait every body */
    barrier_service->barrier(11);

    /* Terminate */
    ACOM::Acom::terminate( );

  } catch (ACOM::Runtime& ex) {
    std::cerr << "Catch exception: " << ex.what() << std::endl;
  }
  catch (...) {
    std::cerr << "Unknown exception" << std::endl;
  }
  return 0;
}



/*
 * Parse options
 */
int Bcast_parseopt( int argc, char* argv[] )
{
#if 1
  optind   =1;
#endif
#if 0
for (int i=0; i<argc; ++i)
{
  std::cout << "argv[" << i << "]=" << argv[i] << std::endl;
}
#endif
  int opt;
  while((opt = getopt(argc, argv, "a:b:hm:v"))>0)
  {
    std::cerr << optarg << std::endl << std::flush;
    switch(opt)
    {
      case 'h':
          Usage();
  
      case 'v':
        verboseon = true;
        std::cout << "Verbose mode on" << std::endl;
      break;
  
      case 'm':
        methodname = optarg;
        if ( methodname == "flat" ) {
        } else if ( methodname == "chain" ) {
        } else if ( methodname == "alpha" ) {
        } else {
          Usage("Bad method");
        }
        if (verboseon)
          std::cout << "Algorithm=" << methodname << std::endl;
      break;
  
      case 'a':
        alpha = atof(optarg);
        if (verboseon)
          std::cout << "Alpha=" << alpha << std::endl;
      break;
  
      case 'b':
        packetsize = atoi(optarg);
        if (verboseon)
          std::cout << "Packet size=" << packetsize << std::endl;
      break;
  
      default:
        ;
    }
  }
  //std::cout << "last index: " << optind << std::endl;
  return optind;
}



/* 
 * Usage about mput1
 */
void Usage(const std::string& msg)
{
  if (msg != "")
    std::cerr << "****** Error: " << msg << std::endl;
  std::cerr << "Usage: mput [-m <method> ] [-b packetsize] [-a <alpha value >] "
            << std::endl;
  std::cerr << "Defaults values: \n"
            << "* method = " << methodname << std::endl
            << "* alpha  = " << alpha << std::endl
            << "* packet size = " << packetsize << " bytes\n"
            << std::endl;
  std::cerr << "Available methods: \n"
            << "* basic algorithms = flat | chain | alpha\n" 
            << std::endl;
  std::cerr << std::endl << std::flush;
  throw;
}



