//
// $Id: rpc0.cpp,v 1.2 2003/07/16 16:48:57 thierry Exp $
// Author : Thierry Gautier  INRIA
// Example of RPC like using Inuktitut :
// - a Thread invokes a service on a remote node and wait 
// the result
// - the handler of the service post a request to a thread
// that reply to the sender


A  F I N I R

#include <acom>
#include <pth>
#include <list>
#include <cstring>
#include <cstdio>

/* fwd declaration*/
class RPCServiceRequester;
class RPC;

// -------------------------------------------------------------------
/* Argument of RPC : a double
*/
typedef double Argument;


// -------------------------------------------------------------------
/* Result of RPC : a double
*/
typedef double Result;


// -------------------------------------------------------------------
/* WaitingContext
   - a flag == true iff the reply has been received
   - the condition to signal
   - an error_code
   - the address where to store the result
*/
struct WaitingContext {
  bool            flag;
  PTH::Condition* cond;
  int             error_code;
  Result*         result;
};


// -------------------------------------------------------------------
/* Remote description of result : 
   - an opaque type that corresponds to a pointer on a waiting context 
   to signal, see above.
   - peer node to reply
*/
struct RemoteWaitingContext {
  unsigned long addr_wc;
  int           peer; 
};


// -------------------------------------------------------------------
/** ServerThread : do service invocation and reply a result
    Should be derived to instanciate a specific server, see below.
    Each ServerThread is a thread associated to a service that waits
    for incomming requests. The service is the entry point to 
    communicate to a remote server. Each ServerThread instance has
    an unique identifier that is used to bind service.
 */
class ClientViewOfServerThread {
protected:
  friend class RPC;
  ACOM::Network*       _net;       /* network where to reply */
  RPCServiceRequester* _service;   /* the service */
};

class ServerThread : public PTH::ThreadModel<ServerThread>, 
                     public ClientViewOfServerThread {
public:
  /** Constructor attached to an inuktitut network
  */
  ServerThread( ACOM::Network* net );
  
  /** Destructor
  */
  ~ServerThread();
  
  /** Type of work processed by such a server
  */
  struct Work {
    Argument             argument;
    Result               result;
    RemoteWaitingContext remote_wc;
  };

  /** Push a RPC call : arguments and where to return result
  */
  void push_request( Work* w );
  
  /** Stop the server
  */
  void stop();
  
protected:
  class DeleteWork : public ACOM::Callback { 
  public:
    /* instanciate work to del */
    DeleteWork( Work* todel );

    /* done method: call when the associated event is done, used to delete a work */
    void done( ACOM::Network::Request* );
  };

  /** entry point of the thread 
  */
  void* run();

  /** Process a request. Should be redefined to implement a specific process.
  */
  virtual void process( Work* const ) =0;

private:
  static int           next_server_id;
  int                  _server_id; /* identifier */
  bool                 _stop;      /* set to true in method stop */
  PTH::Mutex           _access;    /* lock access to internal data members */
  PTH::Condition       _cond;      /* to signal server thread when work arrived */
  PTH::Condition       _condterm;  /* to signal stop is ok */
  std::list<Work*>     _worklist;  /* list of work to process */
};
   

// -------------------------------------------------------------------
/* RPC interface to a Server:
 * - do rpc
 */
class RPC {
public:
  /** Should be called before any use of RPC 
  */
  static void initialize();
  
  /** Should be called after the last use of RPC 
  */
  static void terminate();
  
  /** Build and attach a Client to a RPC server 
  */
  RPC( ClientViewOfServerThread* server );
  
  /** Destroy 
  */
  ~RPC( );
  
  /** Do RPC on node target_node, return value !=0 in case of error
  */
  int invoke( int target_node, const Argument& arg, Result& res );

protected:
  static PTH::Mutex         _access;
  static PTH::Condition     _cond;
  ClientViewOfServerThread* _server;
  WaitingContext            _wc;
};


// -------------------------------------------------------------------
/** Service to push Argument to the server thread
 */
class RPCServiceRequester: public ACOM::ServiceAM {
public:
  /* build and attach a RPCServiceRequester to a ServerThread
  */
  RPCServiceRequester( ServerThread* t );

protected:
  ServerThread* _server;
  
  /* implementation of the service entry point
  */
  void call ( 
	     ACOM::ComFailure::Code e,      // if any 
	     acom::Network::Receipt* receipt,
	     ACOM::Network::Header* h,
	     ACOM::IOVectEntry* iov,
	     ACOM::IOVectSize   iov_sz,
	     void* ctx
	     ) throw (ACOM::ComFailure);
};


// -------------------------------------------------------------------
/** Service to reply result to the client
 *  Message returned by the server is :
 *  address waiting context: <RemoteWaitingContext>, immediate data
 *  error_code             : <int>, immediate data
 *  the data for the result: <Result>, deferred data
 */
class RPCServiceReply: public ACOM::ServiceWS {
public:
  /* build 
  */
  RPCServiceReply( );
  
protected:
  /* implementation of the service entry point
  */
  void call ( 
	     ACOM::ComFailure::Code e,      // if any 
	     acom::Network::Receipt* receipt,
	     ACOM::Network::Header* h,
	     ACOM::IOVectEntry* iov,
	     ACOM::IOVectSize   iov_sz,
	     void* ctx
  ) throw (ACOM::ComFailure);
  
  /* decode address where to write result 
  */
  void* get_iovect ( 
       ACOM::Network::Receipt* r,
       ACOM::Network::Header* h,
       ACOM::IOVectEntry* iov,
       ACOM::IOVectSize   iov_sz
  ) throw (ACOM::ComFailure);
  
};
/* Service reply is unique whatever is the server */
static RPCServiceReply* reply_service;


// -------------------------------------------------------------------
/* Implementation of previous definitions
*/

// -------------------------------------------------------------------
int ServerThread::next_server_id = 1;

/** */
ServerThread::ServerThread(ACOM::Network* net )
 : _server_id(next_server_id), _stop(false)
{
  _access.init();
  _cond.init();
  _condterm.init();
  _worklist.clear();
  _net =net;
  _service = new RPCServiceRequester( this );
  ACOM::Service::Manager.bind (1+_server_id, _service);
}


/** */
ServerThread::~ServerThread()
{
  _access.destroy();
  _cond.destroy();
  _condterm.destroy();
  delete _service;
}


/** */
void ServerThread::push_request( Work* w )
{
  PTH::CriticalGuard critical(_access);
  bool signal = _worklist.empty();
  _worklist.push_back(w);
  if (signal)
    _cond.signal();
}  


/** */
void ServerThread::stop()
{
  PTH::CriticalGuard critical(_access);
  _stop = true;
  _condterm.wait( _access );
}


/** */
void* ServerThread::run()
{
  Work* selected;

  /* Server thread acts as a deamon */
  while (1)
  {
    /* Get work */
    {PTH::CriticalGuard critical(_access);
      if (_worklist.empty()) 
      {
        do {
          _cond.wait(_access);
        } while (_worklist.empty() || _stop);
        if (_stop) {
          return; 
        }
      }
      w = _worklist.front();
      _worklist.pop_front();
    }
    
    /* ok I get a work, process it */
    int error_code = 0;
    try {
      process( w );
    } catch (...) {
      error_code = -1;
    }

    /* Send the result back to the client 
     */
    DeleteWork* dw_callback = new DeleteWork( w );
    ACOM::WriteAndSignal<RPCServiceReply> message( _net, reply_service, dw_callback );
    message->pack(&w->remote_wc, sizeof(RemoteWaitingContext) );
    message->pack(&error_code, sizeof(int) );
    if (error_code ==0)
      message->put(&w->result, sizeof(Result) );
    message.post( _net->get_default_route( w->remote_wc.peer ) );
    
    /* ok goto next request */
  }
}


/** */
ServerThread::DeleteWork::DeleteWork( ServerThread::Work* todel )
 : _work(todel)
{ }


/** */
void ServerThread::DeleteWork::done( ACOM::Network::Request* r )
{
  delete _work;
}


// -------------------------------------------------------------------
/** */
RPCServiceRequester::RPCServiceRequester( ServerThread* t )
 : _server(t)
{ } 


/** */
void RPCServiceRequester::call ( 
      ACOM::ComFailure::Code e,      // if any 
      acom::Network::Receipt* receipt,
      ACOM::Network::Header* h,
      ACOM::IOVectEntry* iov,
      ACOM::IOVectSize   iov_sz,
      void* ctx
      ) throw (ACOM::ComFailure)
{
  /* unpack data in work data structure */
  Work* w = new Work;
  r->unpack( &w->argument, sizeof(Argument) );
  r->unpack( &w->remote_wc, sizeof(RemoteWaitingContext) );
  _server->push_request( w );
}      


// -------------------------------------------------------------------
/** */
RPCServiceReply::RPCServiceReply( )
{
}

/** */
void RPCServiceReply::call ( 
      ACOM::ComFailure::Code e,      // if any 
      acom::Network::Receipt* receipt,
      ACOM::Network::Header* h,
      ACOM::IOVectEntry* iov,
      ACOM::IOVectSize   iov_sz,
      void* ctxt
) throw (ACOM::ComFailure)
{
  WaitingContext* wc  =(WaitingContext*)ctxt;
  wc->flag = true;
  wc->signal();
}

/** */
void* RPCServiceReply::get_iovect ( 
      ACOM::Network::Receipt* r,
      ACOM::Network::Header* h,
      ACOM::IOVectEntry* iov,
      ACOM::IOVectSize   iov_sz
) throw (ACOM::ComFailure)
{
  WaitingContext* wc  =0;
  RemoteWaitingContext rwc;
  r->unpack(&rwc, sizeof(RemoteWaitingContext) );

  /* translate address to recover the WaitingContext 
     and decode the address where to store result
  */
  wc = (WaitingContext*)rwc.addr_wc;
  r->pack(&wc->error_code, sizeof(int) );
  if (error_code ==0)
    iov[0].data = wc->result;
    
  /* return decoded wc to ::call method */
  return wc;
}
  



// -------------------------------------------------------------------
/** */
void RPC::initialize()
{
  _access.init();
  _cond.init();
}


/** */
void RPC::terminate()
{
  _access.destroy();
  _cond.destroy();
}
  

/** */
RPC::RPC( ClientViewOfServerThread* server )
 : _server(server)
{
} 
 
/** */
int RPC::invoke( int target_node, const Argument& arg, Result& res )
{
  _wc.flag   = false;
  _wc.cond   = &_cond;
  _wc.result = &res;
  _wc.error_code = 0;
  RemoteWaitingContext rmt_wc;
  rmt_wc.addr_wc = typeof(rmt_wc.addr_wc)&_wc;
  rmt_wc.peer = _server->_net->get_rank();
  
  /* Form the message */
  ACOM::ActiveMessage<RPCServiceRequester> message(_server->_net, _server->_service, 0);
  message->pack( &arg, sizeof(Argument) );
  message->pack( &rmt_wc, sizeof(RemoteWaitingContext) );
  message.post( _server->_net->get_default_route(target_node) );
  PTH::CriticalGuard critical(_access);
  while (!_wc.flag) 
    _cond.wait(_access );
  return _wc.error_code;
}


// -------------------------------------------------------------------
/* User class that compute the square root of a double value
*/
class ServerSquareRoot : public ServerThread {
protected:
  void process( ServerThread::Work* w )
  {
    if (w->argument <0) throw;
    w->result = sqrt(w->argument);
  }
};


// -------------------------------------------------------------------
int doit (int argc, char* argv[])
{
  ACOM::Acom::initialize( &argc, &argv );
  
  /* declare and bind the service */
  reply_service = new RPCServiceReply;
  ACOM::Service::Manager.bind (1, reply_service);

  ACOM::Acom::commit( );


	/* Get the default network:
  */
  ACOM::Network* net = ACOM::Acom::get_default_network();
  int rank = net->rank();
  
  /* Create and start the thread Server
  */
  ServerSquareRoot do_sqrt;
  do_sqrt.initialize( net );
  do_sqrt.start();
  
  /* wait all server start */
  net->barrier();
  
  /* Make a call from 0 to 1 */
  if (rank ==0) {
    RPC rpc( &do_sqrt )
    Argument a = 4; 
    Result   r = 0; 
    int err = rpc.invoke( 1, a, r );
    std::cout << "Error code is =" << err;
    if (err ==0) std::cout << ", result of sqrt(" << a << ") is:" << r;
    std::cout << std::endl;

    a = -1; 
    r = 0; 
    int err = rpc.invoke( 1, a, r );
    std::cout << "Error code is =" << err;
    if (err ==0) std::cout << ", result of sqrt(" << a << ") is:" << r;
    std::cout << std::endl;
  }
  
  /* wait all client make call */
  net->barrier();
  do_sqrt.stop();
  delete reply_service;

  ACOM::Acom::terminate( );
}


// -------------------------------------------------------------------
/** Main entry point
*/
int main (int argc, char* argv[])
{
  try {
    return doit(argc, argv);
  } catch (...){
    std::cerr << "Unknown exception" << std::endl;
  }
}
