//
// $Id: testcc2.cpp,v 1.2 2003/07/28 07:40:11 thierry Exp $
// Author : T Gautier,  INRIA
// test active message: 
//
#include <acom>
#include <pth>
#include <list>
#include <cstring>
#include <cstdio>
#include <unistd.h>
#include <sys/time.h>
#include <iostream>
#include <iomanip>


static int SIZE_PACKET  = 1;
static int COUNT_PACKET = 1;


/* The network used for all communications
*/
static ACOM::Network* net = 0;
static int rank = 0;
static int nodecount = 0;

/* ===========================================================================
 * Definition the services and thread's class for redirection
 * =========================================================================== */

/* Ack service that send ack to node 0
*/
class AckService: public ACOM::ServiceCC {
public:
  /* Cstor */
  AckService();
  
  /* Dstor */
  ~AckService();

  /* wait ack */
  void wait();

protected:
  PTH::Condition _cond;
  PTH::Mutex     _access;
  bool           _ok;
  
  /* This is low level entry point of a service : 
  */
  void server_call( ACOM::Network::CallContext* cc ) 
    throw(ACOM::ComFailure);
};
static AckService* ack_service;


/* Echo service : display message and push cc to the thread for
   redirection 
*/
class EchoService: public ACOM::ServiceCC {
public:
  /* Cstor */
  EchoService();
  
  /* Dstor */
  ~EchoService();

protected:
  int        _count_called;  /* count the number of called */
  PTH::Mutex _access;

  /* This is low level entry point of a service : 
  */
  void server_call( ACOM::Network::CallContext* cc ) 
    throw(ACOM::ComFailure);
};
static EchoService* echo_service;


/* thread that handles list of CallContext to redirect them to next node
*/
class ThreadRedirec {
public:
  /* build a thread that will redirect a message to dest nn */
  ThreadRedirec( int nn );
  
  /* destroy the thread objet */
  ~ThreadRedirec( );
  
  /* Ping service push a new cc to redirect, that will
     be re-send by the current thread.
     Push in back, pop in front to conserve fifo
  */
  void push( ACOM::Network::CallContext* cc );
  
  /* Consumer of cc, entry point of the thread */
  void run();
  
  /* Stop the thread */
  void stop();
  
public:
public:
  /* entry point for PTH thread creation */
  static void* run(void* arg);

protected:
  PTH::Condition                         _cond;
  PTH::Condition                         _condstop;
  PTH::Mutex                             _access;  
  int                                    _nextnode;     /* next node where to redirec */
  bool                                   _wait;         /* true if the thread wait a cc */
  bool                                   _stop;         /* true if the thread should stop */
  std::list<ACOM::Network::CallContext*> _cc_toredirec; /* list of cc to redirect */  
};
static ThreadRedirec* thread;  



/* ===========================================================================
 * Implantation of previous methods
 * =========================================================================== */

/*
*/
AckService::AckService()
{
  _cond.init();
  _access.init();
  _ok = false;
}
  

/*
*/
AckService::~AckService()
{
  _cond.destroy();
  _access.destroy();
}


/*
*/
void AckService::wait()
{
  PTH::CriticalGuard critical( _access );
  while (!_ok)
    _cond.wait( _access ); 
  _ok = false;
}


/*
*/
void AckService::server_call( ACOM::Network::CallContext* cc ) 
  throw(ACOM::ComFailure)
{
  PTH::CriticalGuard critical( _access );
  std::cout << "In ack service" << std::endl;
  _ok = true;
  _cond.signal();
}


/*
*/
EchoService::EchoService()
 : _count_called(0)
{
  _access.init();
}
  

/*
*/
EchoService::~EchoService()
{
  _access.destroy();
}


/*
*/
void EchoService::server_call( ACOM::Network::CallContext* cc ) 
  throw(ACOM::ComFailure)
{
  if (cc->error_no != ACOM::ComFailure::OK) 
  {
    std::cerr << "Error in msg='" << cc->error_no << std::endl;
  }
  
  bool called_ack = false;
  { PTH::CriticalGuard critical( _access );
    ++_count_called;
    called_ack = (_count_called == COUNT_PACKET);
    std::cout << "In echo service: count =" << _count_called << std::endl;
  }

  if ((rank + 1) % nodecount != 0) {
    thread->push( cc );
  }
  else if (called_ack) 
  {
    /* final node : send ack to */
    ACOM::Network::OutChannel* channel = net->get_default_route(0);
    ACOM::ActiveMessage<AckService> am (net, ack_service, 0);
    am.post(channel);
    _count_called =0;
  }
}


/* build a thread that will redirect a message to dest nn
*/
ThreadRedirec::ThreadRedirec( int nn )
 : _nextnode(nn)
{
  _access.init();
  _cond.init();
  _condstop.init();
  _wait = false;
  _stop = false;
  _cc_toredirec.clear();
}


/*
*/
ThreadRedirec::~ThreadRedirec( )
{
  _access.destroy();
  _cond.destroy();
  _condstop.destroy();
  _cc_toredirec.clear();
}


/*
*/
void ThreadRedirec::stop()
{
  PTH::CriticalGuard critical(_access);
  _stop = true;
  _cond.signal();
  while (_stop) {
    _condstop.wait(_access);
  }
}


/*
*/
void ThreadRedirec::push( ACOM::Network::CallContext* cc )
{
  PTH::CriticalGuard critical(_access);
  _cc_toredirec.push_back( cc );
  cc->acquire();
  if (_wait) _cond.signal();
}

  
/* 
*/
void ThreadRedirec::run()
{
  /* redirect the cc to _nextnode */
  ACOM::Network::OutChannel* channel = net->get_default_route(_nextnode );
  ACOM::Network::CallContext* cc = 0;
  
  int count = 0;
  
  /* this is a daemon's like thread, infinite loop except if it stop */
  while (1)
  {
    /* get a cc to redirect */
    { PTH::CriticalGuard critical(_access);
      if (_cc_toredirec.empty()) 
      {
        _wait = true;
        while (_cc_toredirec.empty() && !_stop) 
          _cond.wait( _access );
        _wait = false;
        if (_cc_toredirec.empty() && _stop) return;
      }
      cc = _cc_toredirec.front();
      _cc_toredirec.pop_front();
    }
    
    int sz =0;
    char* banner =0;
    cc->receipt.unpack( &sz, sizeof(int) );
    banner = (char*)::alloca( sz );
    cc->receipt.unpack( banner, sz );
    std::cout << rank << "::" << banner << ++count << std::endl;
  
    /* ok get a cc, resend it to _nextnode */
    ACOM::ActiveMessage<EchoService> am (net, echo_service, 0);
    am->pack( &sz, sizeof(int) );
    am->pack( banner, sz );
    
    /* for all deferred data, reform a message */
    for (int i=0; i< cc->iov_sz; ++i)
      am->put(&cc->iov[i]);
      
    /* post the message */
    am.post( channel );
    
    /* release the cc */
    cc->release();
  }
}


/* 
*/
void* ThreadRedirec::run(void* arg) 
{
  ThreadRedirec* obj = (ThreadRedirec*)arg;
  try {
    obj->run();
    PTH::CriticalGuard critical(obj->_access);
    obj->_stop = false;
    obj->_condstop.signal();
  } catch (const ACOM::Runtime& ex) {
    std::cout << "Catch error= " << ex.what() << std::endl;
  }
  return 0;
}
  


/* ===========================================================================
 * Do it method
 * =========================================================================== */
int doit ()
{
  if (rank ==0) 
  {
    /* Get a channel, form the message
    */
    char* data = new char[SIZE_PACKET];
    ACOM::Network::OutChannel* channel = net->get_default_route(1);
    ACOM::ActiveMessage<EchoService> am (net, echo_service, 0, COUNT_PACKET);
    char* banner = "Packet,  packet, ...";
    int sz = strlen(banner)+1;
    am->pack( &sz, sizeof(int) );
    am->pack( banner, sz );
    am->put(data, SIZE_PACKET);

    for (int k=0; k < COUNT_PACKET; ++k) {
      am.post(channel);
    }
    ack_service->wait();
  }
  net->barrier();
}


/* ===========================================================================
 * Main
 * =========================================================================== */
int main (int argc, char* argv[])
{
  try {
    /* Initialize network */
    ACOM::Acom::initialize( &argc, &argv );
    
    /* create the service objects */
    ack_service = new AckService;
    echo_service = new EchoService;
    ACOM::Service::Manager.bind (2, ack_service);
    ACOM::Service::Manager.bind (3, echo_service);

    /* Commit network */
    ACOM::Acom::commit( );

    /* Get the network:
    */
    net = ACOM::Acom::get_default_network();

    /* Wait every body */
    net->barrier();
    
    /* rank and nodecount */
    rank = net->get_rank();
    nodecount = net->get_nodecount();

    /* create the redirection thread */
    thread = 0;
    if ((rank+1) % nodecount !=0) 
    {
      thread = new ThreadRedirec((rank+1) % nodecount) ;
      PTH::ThreadId tid;
      PTH::Thread::create( &tid, &ThreadRedirec::run, (void*)thread);
    }
    net->barrier();

    /* Get the argument
    */
    if (rank ==0) 
    {
      std::cout << "Network is=" << net->get_name()
                << ",  my rank is=" << rank 
                << ", over #nodes=" << nodecount << std::endl;
    }
    SIZE_PACKET  = atoi(argv[1]);
    COUNT_PACKET = atoi(argv[2]);
    
    std::cout << "SIZE_PACKET=" << SIZE_PACKET << std::endl;
    std::cout << "COUNT_PACKET=" << COUNT_PACKET << std::endl;

    net->barrier();

    /* Send : COUNT_PACKET of size SIZE_PACKET */
    doit();
    if (thread !=0)
      thread->stop();
    net->barrier();

    /* Terminate */
    ACOM::Acom::terminate( );

    /* Delete services */
    delete echo_service;
    delete ack_service;

  } catch (ACOM::Runtime& ex) {
    std::cerr << "Catch exception: " << ex.what() << std::endl;
  }
  catch (...) {
    std::cerr << "Unknown exception" << std::endl;
  }
  return 0;
}
