// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/Pthread/src/pth_pthreadsync.h,v $ 
//Copyright(c)'2001 by Apache Team // see the copyright file 
// $Id: pth_pthreadsync.h,v 1.9 2003/07/28 23:30:55 thierry Exp $ 
//==========================================================================

/*!
\file  
\brief  Posix thread for Inuktitut0++ INRIA 
\author  J. Briat T. Gautier 
\date 2001

This file contains the wrappers to :
  - posix locks PTH::Mutex
  - posix condition PTH::Condition
  - part of the posix thread interface PTH::Thread PTH::ThreadModel
  - a lexical scoped lock mechanism CRITICAL()

It provides operators on these objects throwing Posix Exceptions.
\sa pth_error.h PTH::PosixError
*/
#ifndef _PTH_PTHREADSYNC_H
#define _PTH_PTHREADSYNC_H

// the namespace identifier
#include "pth_ns.h"
#include "pth_dep.h"
#include "pth_macro.h"
#include "pth_error.h"
#include <sys/time.h>


#if defined(PTH_USE_PTHREAD_DARWIN) || defined(PTH_USE_PTHREAD_LINUX) || defined(PTH_USE_PTHREAD_IRIX)
#include <pthread.h>
#endif 

#if defined(PTH_USE_PTHREAD_DARWIN) 
#define MUTEX_TYPE_NORMAL PTHREAD_MUTEX_NORMAL
#define MUTEX_TYPE_RECURSIVE PTHREAD_MUTEX_RECURSIVE

#elif defined(PTH_USE_PTHREAD_LINUX)
#define MUTEX_TYPE_NORMAL PTHREAD_MUTEX_FAST_NP
#define MUTEX_TYPE_RECURSIVE PTHREAD_MUTEX_RECURSIVE_NP

#elif defined(PTH_USE_PTHREAD_IRIX)
#define MUTEX_TYPE_NORMAL PTHREAD_MUTEX_NORMAL
//#define MUTEX_TYPE_NORMAL PTHREAD_MUTEX_SPINBLOCK_NP
#define MUTEX_TYPE_RECURSIVE PTHREAD_MUTEX_RECURSIVE
//#define COND_TYPE PTHREAD_PROCESS_SHARED
#define COND_TYPE PTHREAD_PROCESS_PRIVATE
#else
#error UNABLE TO DEFINE POSIX THREAD ATTRIBUTES
#endif

namespace PTH { 

typedef pthread_mutex_t PMutex;
typedef pthread_cond_t  PCondition;

class Pth; // forward decl
// -------------------------------------------------------------------------
// classes Mutex
// -------------------------------------------------------------------------
/**
\brief This class implements interface for Posix mutex.
Be careful to unlock a mutex in case of exception throwing

Use class CriticalGuard or macro CRITICAL to do this automatically
\code
#include "pth.h"
// - declaration 
static Mutex my_mutex;
// --exception unsafe version, 
void foo1() {
  // here do something
  
  { // I want this block to be executed in exclusive access
    my_mutex.lock();
    ....
    ....
    my_mutex.unlock();
  } 
  // here do something
 }
void foo2() {
  // here do something
  
  { // I want this block to be executed in exclusive access
    CriticalGuard gd(my_mutex);
    ....
    ....
    
  } // end of exclusion
  // here do something
}
\endcode
\sa pth::CriticalGuard 
*/

class Mutex {
public:
  /// Constructor for a new uninitialized mutex in unlock state.
  Mutex() throw () 
#ifdef PTH_DEBUG
   : _isinit(false)  
#endif
  {}

  /// Initialize a  recursive or flat (default case) mutex in unlock state.
  void init(bool recursive=false) throw (PosixError);
  
  /// virtual destructor
  virtual ~Mutex() throw (PosixError) {
#ifdef PTH_DEBUG
    _isinit =false;  
#endif
  } 
  
  /// Explicit destructor. \throw for an unitialized mutex \throw for a locked mutex.
  void destroy () throw (PosixError);
  /// lock request. Wait until access is granted. \throw for an unitialized mutex
  void lock() throw( PosixError); 	 
  /// lock if possible (true case). Wait until access is granted.  \throw for an unitialized mutex
  bool trylock() throw (PosixError);
  ///  unlock request. \throw  for an unitialized mutex \throw if the current thread is not owner.
  void unlock() throw( PosixError);

  /// @name Operators returning an error code
  //@{
  /// Initialize a  recursive or flat (default case) mutex in unlock state.(For compatibility with previous release )
  int initnothrow(bool recursive = true ) throw ();
  /// Explicit destructor. \return  0 if OK else n#0 for an unitialized mutex or a locked mutex. 
  int destroynothrow () throw ();
  /// lock request. \return 0 if OK else n#0 for an unitialized mutex . 
  int locknothrow() throw ();
  ///  unlock request.	\return 0 if OK else n#0 for an unitialized mutex or if the current thread is not owner.
  int unlocknothrow() throw ();
  /// lock if possible. \return 0 if lock successful , EBUSY if lock failed else #(0 or EBUSY) for an unitialized mutex   
  int trylocknothrow() throw (); 
  //@}

  /// @name Used in  tracer only. Avoid to trace the tracer !!!
  //@{
  void init_notrace() throw ();   
  void lock_notrace() throw( PosixError);  
  void unlock_notrace() throw( PosixError);  
  //@}

protected: 
  /// the posix lock
  pthread_mutex_t _mutex;
#ifdef PTH_DEBUG
  bool _isinit;
#endif
private :
  static  pthread_mutexattr_t default_attr;
  static  pthread_mutexattr_t recursive_attr;
  /// static values initialisation 
  static void init_static() throw (PosixError); 
  
  friend class Condition;
  friend class Pth;
};


// -------------------------------------------------------------------------
// classes Condition
// -------------------------------------------------------------------------
/** 
\brief This class implements interface to posix condition variable.
*/

class Condition {
public:   
  /// Constructor
  Condition() throw ()
  {}
  ///   Initialize a  condition.
  void init() throw (PosixError); 
  ///  Explicit destroy ; Throws an exception if there is waiting threads.
  void destroy () throw (PosixError);

  ///Destructor. Throws an exception if there is waiting threads.
  virtual ~Condition() throw (PosixError) 
  { } 

  /**  Wait :release lock and wait on condition. Relock mutex when waked up. 
    \throw for  unitialized mutex \throw for  unitialized cond  \throw if not owned mutex..
  */
  void wait( Mutex& mutex ) throw (PosixError);

  /** \brief Bounded Wait :release the lock and wait for a bounded time 
    (in second/microsecond) on the condition. Relock when waked up. 
    \return true if awaked by timeout 
    \throw for  unitialized mutex \throw for  unitialized cond \throw for  a not owned mutex.
  */
  bool wait( Mutex& mutex , long milli, int nano=0) throw(PosixError); 

  /** \brief Bounded Wait :release the lock and wait for a bounded time 
    (on Unix timespec coding) on the condition. Relock when waked up. 
    \return true if awaked by timeout  
    \throw for  unitialized mutex 
    \throw for  unitialized cond 
    \throw for a not owned mutex.
  */
  bool wait( Mutex& mutex, const struct timespec& abstime ) throw (PosixError);
  /// Signal:awake ONE waiting thread only. \throw for  unitialized  cond 
  void signal() throw (PosixError);
  ///  Broadcast :awake ALL waiting thread . \throw for  unitialized  cond
  void broadcast() throw (PosixError); 
    
  /// @name Operators returning an error code
  //@{
  ///   Initialize a  condition.(Compatibility with previous versions)
  int initnothrow() throw ();
  ///  Explicit destroy ; \return #0n if there is waiting threads.
  int destroynothrow () throw ();
  /**  Wait :release lock and wait on condition. Relock mutex when waked up. 
    \return #0 for  unitialized mutex or cond or not owned mutex..
  */
  int waitnothrow( Mutex& mutex ) throw ();

  /** Bounded Wait :release the lock and wait for a bounded time on the condition. Relock when waked up. 
  \return ETIMEOUT if awaked by timeout. 
  \return #(0 or ETIMEOUT) for unitialized mutex or cond or a not owned mutex..
  */
  int waitnothrow( Mutex& mutex ,long milli, int nano=0) throw ();

  /** Bounded Wait :release the lock and wait for a bounded time 
  (on Unix timespec coding) on the condition. Relock when waked up. 
  \return ETIMEOUT if awaked by timeout. 
  \return #(0 or ETIMEOUT) for unitialized mutex or cond or a not owned mutex..
  */
  int waitnothrow( Mutex& mutex , const struct timespec& abstime ) throw ();
  /// Signal:awake ONE waiting thread only. \return 0 if OK else #0  unitialized  cond
  int signalnothrow() throw ();
  ///  Broadcast :awake ALL waiting thread .\return 0 if OK else #0  unitialized  cond
  int broadcastnothrow() throw ();  
  //@}

protected:
  /// the posix condition
  pthread_cond_t _condition;
  
private:
  friend class Pth;
  static pthread_condattr_t Condition::default_attr; 
  static void init_static() throw (PosixError); 
  // friend void initialize(int* , char***) throw (PosixError);
};


// -------------------------------------------------------------------------
// class CriticalGuard
// -------------------------------------------------------------------------
/** 
\brief This class implement a  transparent local scoping usage of mutex lock.
An instance of this class is used to lock a block expression (i.e. between { and }).
The instance delaration must appear at the beginning of the block.
At the initialization of such an object, the associated mutex is locked and 
is unlocked at the destruction of the object (at the end of its scope as a 
C++ variable). Note that the unlock method is alway called even if an exception 
is thrown.
\code
#include "pth.h"

// - declaration but not initialization
static Mutex my_mutex;

void dummy1() {
  // here do something

  { // I want this block to be executed in exclusive access
     CriticalGuard guard(my_mutex);
    ....
    ....
  } // here the mutex is released when "guard" is destroyed ..

  // here do something
}
\endcode

It is possible to use the macro CRITICAL(). 
*/
class CriticalGuard {
private :
  PTH::Mutex * _mtx_;
public : 
  CriticalGuard (Mutex &mtx) : _mtx_(&mtx) {  _mtx_->lock();};
  CriticalGuard (Mutex *mtx) : _mtx_(mtx) {  _mtx_->lock();};
  ~CriticalGuard () { _mtx_->unlock();};
};


/** 
\brief This class implement a  transparent local scoping usage of mutex lock.
 Its only used in nthe tracing facility to avoid to trace the tracer critical sections. \sa CriticalGuard for the meaning 
*/
// without trace version of CriticalGuard 
class CriticalGuard_notrace {
private:
    PTH::Mutex * _mtx_;
public: 
  CriticalGuard_notrace (Mutex &mtx) : _mtx_(&mtx) {  _mtx_->lock_notrace();};
  CriticalGuard_notrace (Mutex *mtx) : _mtx_(mtx) {  _mtx_->lock_notrace();};
  ~CriticalGuard_notrace () { _mtx_->unlock_notrace();};
};


/**
\def  CRITICAL()
\brief This macro can be use in place of PTH::CriticalGuard 
*/
#define CRITICAL(MTX) PTH::CriticalGuard  _CONCAT(_cg, __LINE__) (MTX)


/**
\def  CRITICAL_NOTRACE()
\brief This macro is used in the tracing facility in place of  CRITICAL() to avoid to trace the the tracer critical sections.
*/
#define CRITICAL_NOTRACE(MTX) PTH::CriticalGuard_notrace  _CONCAT(_cg, __LINE__) (MTX)



// -------------------------------------------------------------------------
// class Key<T>
// -------------------------------------------------------------------------
/** 
\brief This class Key provides an  interface to posix pthread specific
attributes needed for  Inuktitut0 INRIA. threads

It only includes operators for :
  - getting specific data key create() with a value destructor function called at the end of the thread
  - setting/getting specific data set() get()

*/
template < class T> class Key{
protected :
  pthread_key_t key;
public :
  Key() throw()
  {}
  ~Key() throw()
  {}

  /// allocate a key. \return 0 if OK and #0 if nomore keys
  int createnothrow(void (*erase)(T*)=0) throw ()
  { return  pthread_key_create(&key,(void (*)(void*))erase ); }
  /// allocate a key. \throw  if nomore keys
  void create(void (*erase)(T*)=0) throw (PosixError)
  { int err; 
    ASSERT( !(err=pthread_key_create(&key,(void (*)(void*))erase)), 
      PosixError (SUFX("[Key::signal] POSIX error"),err));
  }

  /// get the value associated with a given key
  T* get() throw ()
  { return (T*) pthread_getspecific(key); }
  
  /// associate  a value to a key. \return 0 if OK #0 si bad key
  int setnothrow( T * v) throw ()
  { return pthread_setspecific(key, (void *) v); }
  /// associate  a value to a key. \throw when bad key
  void set( T * v) throw (PosixError)
  { int err; 
    ASSERT( !(err=pthread_setspecific(key, (void *) v)), 
      PosixError (SUFX("[Key::signal] POSIX error"),err));
  }
};   



#define PTH_CLEANER(HH,AA) PTH::Cleaner _CL_(HH,AA)
/** 
\brief This class Cleaner provides an  interface to posix cleanup handler
    needed for  Inuktitut0  threads
CleanUp handlers MUST be allocated on stack and MUST have block or procedure scope
*/  
class Cleaner {
protected :
#if defined(PTH_USE_PTHREAD_LINUX)
  struct _pthread_cleanup_buffer _clup;
#elif defined(PTH_USE_PTHREAD_DARWIN)
  struct _pthread_handler_rec __handler;
#endif
  // TODO :
public :
  Cleaner(void (*hdl)(void*), void*  v);
  ~Cleaner();
};

typedef pthread_t ThreadId;

#define PTH_DISABLE_CANCEL  PTH::Thread::disableCancel()

/** 
\brief These functions provide an  interface to posix pthread 
  needed for  Inuktitut0 threads. 
It only includes operators for :
  - cancellation disabling  Thread::disableCancel( )
  - yielding control Thread::yield()
  - sleep during a period of time Thread::sleep()
  - creating Pthread in joinable or detached mode Thread::create() Thread::createnothrow()
	- joining pthread started in joinable mode Thread::join() Thread::joinnothrow()
  - getting the current thread id Thread::getSelf()
*/
class Thread {
public:
  /// attribute type of a thread
  typedef pthread_attr_t * Kind;
  /// To create a thread as detached
  static  Kind DETACHED;
  /// To create a thread as joinable
  static  Kind JOINABLE;
  Thread() throw() {}
  ~Thread() throw() {}
  
  /// current thread yield the processor if its priority is lower than ready threads
  static void yield() throw ();
  /// current thread yield the processor for an given delay
  static int sleep( long milli, long nano) throw ();
  /// current thread disable automatic cancellation
  static  int disableCancel( ) throw();
  /// terminate the thread with a pointer on a return value ( posix type) if exists
  static void exit( void*x=0)throw (){ pthread_exit(x);};
  /// terminate the thread with a pointer on a value ( cast to posix type) if exists
  template <class Tr >
  static void exit( Tr*x=0)throw (){ pthread_exit((void*)x);};


  /// @name  Thread creation
  //@{
  ///create a detached thread(posix type). An  error code is returned if it fails
  static int createnothrow(ThreadId * thid, void  *(*f)(void *),void * a ,Kind k= DETACHED ) 
     throw();
  ///create a detached thread ( cast argument to posix type). An  error code is returned if it fails
  template <class Tr , class Ta> 
  static int createnothrow(ThreadId * th, Tr *(*f)(Ta *), Ta* a ,Kind k= DETACHED) 
     throw()
  { return createnothrow(th, (void  *(*)(void *))f , (void *)a, k);} ;
  /// create a detached thread (posix type). An exception is thrown if it fails
  static void create(ThreadId * thid, void  *(*f)(void *),void * a, Kind k= DETACHED) 
     throw(PosixError);
  /// create a detached thread ( cast argument to posix type) An exception is thrown if it fails
  template <class Tr , class Ta> 
  static void create(ThreadId * thid, Tr *(*f)(Ta *),Ta* a, Kind k= DETACHED ) 
    throw(PosixError)
  { create(th, (void  *(*)(void *))f , (void *)a, k); };
  //@}


  /// @name  Thread junction.  Get a pointer on a result value if second param  is not 0
  //@{
  /// join a joinable thread. (posix type).An  error code is returned if it fails
  static int joinnothrow(ThreadId  thid,void **v=0) throw();
  /// join a joinable thread.( cast argument to posix type). An  error code is returned if it fails
  template <class Tr >
  static int joinnothrow(ThreadId  thid, Tr** v=0)
  { return joinnothrow( thid,(void**)v); }
  /// join a joinable thread. (posix type). An exception is thrown if it fails
  static void join(ThreadId  thid,void **v=0) throw (PosixError);
  ///join a joinable thread.( cast argument to posix type). aN EXCEPTION is thrown if it fails.
  template <class Tr >
  static int join(ThreadId  thid, Tr** v=0)
  { join( thid,(void**)v); }
  //@}

  /// get current thead id thread   
  static ThreadId  getSelf()throw ()
  { return ::pthread_self(); }

private:
  static  pthread_attr_t detached_attr;
  static  pthread_attr_t joinable_attr;
  /// static values initialisation 
  static void init_static() throw (PosixError); 
  friend class Pth;
};


/** 
\brief These class  provide a wrapper class fromC++ object to  posix pthread 
  needed for  Inuktitut0 threads .

Any classes providing a void run() method may be executed as a thread.

It inherits of the Thread class operators. It  includes  operators :
	- to start a thread and encapsulate its id. PTH::ThreadModel::start() PTH::ThreadModel::startnothrow()
	- to join a thread PTH::ThreadModel::join() PTH::ThreadModel::joinnothrow()
	- to get the Posix thread id PTH::ThreadModel::getThreadId()
*/
template < class XX > 
class ThreadModel : public Thread {
static void* runIt( void *s);
public:
  /* Need */
  virtual ~ThreadModel() {}

  /* Needed to make the right call to object XX
     to validate translation : 
     - start: ThreadModel* -> XX*
     - runIt: XX* a, a->run
     Cost a virtual method call
  */
  virtual void* run() =0;
  
  /// Native thread id . Is it  really used ??
  ThreadId       _threadid;    

  /// start a thread object. An exception is thrown if it fails
  void start(Thread::Kind k= Thread::DETACHED) throw(PosixError)
  { 
    //std::cout << "Start: @obj =" << (void*)this << std::endl;
    create(& _threadid, (void *(*)(void*))runIt , this, k ); 
  }

  /// start a thread object. An  error code is returned if it fails
  int startnothrow(Thread::Kind k= Thread::DETACHED) throw(PosixError)
  { return create(&_threadid, (void *(*)(void*))runIt, this , k) ; }

  /// join the thread. An exception is thrown if it fails. Get a pointer on the result value 
  template < class Tr >
  void join( Tr** v=0) throw (PosixError)
  { Thread::join(_threadid,&v); }

  /// join the thread. An  error code is returned if it fails. Get a pointer on the result value
  template < class Tr >
  int joinnothrow( Tr** v=0) throw ()
  { Thread::join(_threadid,v); }

  /// get native  thread id
  ThreadId getThreadId()
  { return _threadid; }
};

//class ClassicThread : public ThreadModel< ClassicThread >{};

//ClassicThread CL;

}// -namespace
#endif // _PTH_PTHREADSYNC_H_
