// =========================================================================		
// $Source: /users/huron/gautier/CVS/Inuktitut2/Pthread/src/tr_recorder.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: J. Briat & T. Gautier

//
// ==========================================================================
#ifndef _TR_RECORDER_H_
#define _TR_RECORDER_H_
/*!
  \file
  \brief Inuktitut INRIA Recorder
  \author  J. Briat 

  This file is dedicated to the buffer and event recording management classes for Inuktitut INRIA tracing facility

  This class implement a mechanism. All event are completely recorded in a buffer. When buffer cannot hold an effet, the buffer is flushed in a trace file and a fresh buffer is used for the event.  
	
  When flushing a buffer, a  flush event record is inserted in the new buffer. This will be used to trace the behavior of the buffer allocation strategy.

  Recording of value is done by memory block copy in the buffer using the templated operator <<.
  This template must be specialized for specific needs. This is the case for the std::string type.

  This part is environnement dependant and must be adapted as needed.

  \date 2002
*/

 

#include "tr_buffer.h"
#include "pth_system.h"
#include "pth_pthreadsync.h"
#include "pth_atomic.h"
#include "pth_basic.h"
#include <list>
#include <string>


namespace PTH 
{
class  SimpleRecorder ; // forward ref
  
  /** 
      \brief  A Generic Trace Buffer class : PTH::SimpleBuffer
      
      This class implement the recording of value in buffer. 
      Recording of value is done by memory block copy in 
      the buffer using the templated operator <<.
      This template must be specialized for specific needs. 
      This is the case for the std::string type.
  */
  
  class  SimpleBuffer :public Queable<SimpleBuffer > {
      public :
/// Buffer size
	enum Size { SZ=4096}; // event zone size
/// Associated thread
      PTH::ThreadId _tid; //associated thread
/// scan pointer for writing
      char* _scan;
/// filled size
      int _sz; // filled size
/// buffer
      char _data[SZ]; // events zone:
      public :
	/// constructor
	SimpleBuffer():_tid(0){_sz=SZ;_scan= &_data[0];};
  /// destructor
  ~SimpleBuffer(){}
  /// open a trace buffer
      void open(){_sz=SZ;_scan= &_data[0];}
	/// reset the trace buffer 
      void refresh(){_sz=SZ;_scan= &_data[0]; }  
  /// close thre trace buffer
      int close (){return SZ-_sz; }
 };
  
/// register part of an event in the buffer 
//@{
  /// register byte block. This template may be specialized as needed
  template < class T >
    inline  SimpleBuffer &operator<<(SimpleBuffer &b,T t)
    {
      memcpy(b._scan,&t,sizeof(T));
      b._sz-=sizeof(T);
      b._scan+=sizeof(T);
      return b;
    }
/// specialization  for string 
  inline  SimpleBuffer & operator<<(SimpleBuffer &b,std::string t)
    {
      // recording of string's size
      StringSize sz=t.size();
      memcpy(b._scan, &sz, sizeof(StringSize));	
      b._sz -= sizeof(StringSize);
      b._scan+=sizeof(StringSize);
      //copy the string
      t.copy(b._scan, sz);
      b._sz -= sz;
      b._scan+=sz;
      return b;
    }
//@}
  /** \brief  A Generic Buffer Manager : PTH::SimpleBufferManager

  SimpleBufferManager is an implementation of a buffer manager for tracing:
  -  blocking trace when saving a buffer on disk	
  -  using an unique file to dump buffers

  */


class SimpleBufferManager {   
protected:
  //----------------------list_buffer---------------------------

	/// list of trace buffer
	Queue<SimpleBuffer> list_buffer;  
	/// mutex for the buffer list  
	PTH::Mutex mutex_list_buffer;

 //--------------------------------------------------------------
	/// mutex for trace file
	PTH::Mutex mutex_fd;
	/// trace file
	int _fd; // file descriptor
public:
/// constructor
	SimpleBufferManager():_fd(-1){};
/// desstructor
	~SimpleBufferManager(){};
/// Allocate  a buffer
	SimpleBuffer * allocBuffer( );
/// free  a buffer
	void freeBuffer( SimpleBuffer *);
/// save &free  all  buffers
	void saveFreeAllBuffer( );
/// . Save buffer if needed
	void saveBuffer( SimpleBuffer * tr);
///  Save buffer and provide another fresh buffer
	SimpleBuffer * flushBuffer(SimpleBuffer * tr);
/// initialize manager. Open file etc...
	void openTraceFile(   char *fn  )throw(PosixError);
/// terminate manager. Close file etc...
	void  closeTraceFile()throw(PosixError);
/// init
 void  init(){ 
  //----------------------list_buffer---------------------------
	mutex_list_buffer.init_notrace();
//--------------------------------------------------------------
	mutex_fd.init_notrace();}
    };
 
  
  
  /** 
      \brief  Thread aware generic Binder  : PTH::Binder
      
      maintains association betwwen the current thread and its tracer  ( and the inverse)
      
  */
  
  class Binder{
/// key used to link a thread to it trace buffer
	PTH::Key<SimpleBuffer> _tracekey;
public :
/// Constructor
	Binder(){};
/// Destructor
	~Binder(){};
/// get the buffer bound to the current thread
	SimpleBuffer * getBuffer(){ return _tracekey.get(); }
/// bind a buffer to the current thread
	void bindBuffer(SimpleBuffer* tr){
	  _tracekey.setnothrow(tr);
	  tr->_tid= Thread::getSelf();
	}
/// unbind a buffer bound to the current thread
	SimpleBuffer * unBindBuffer(){
	  SimpleBuffer* tr  = _tracekey.get();
	  _tracekey.setnothrow(0);tr->_tid= 0;
	  return tr;
	}
/// initialize binder. 
      void initBinder(void (*destruct)(SimpleBuffer*)=0) throw(PosixError);
  };


  /** 
      \brief  A simple Recorder class : PTH::SimpleRecorder

      This class implement a  trace buffer mechanism.  It inherits from :
      - a buffer to be used for trace
      - a buffer management class ( inherited) The flushing of a buffer is done by a buffer manager which  manages the buffer allocation and saving on disk.
      - the binder to get the current trace buffer associated to a thread(inherited) 
      - the clock and time representation to be used to timestamp  event record

      A generic recorder is associated with a portable level thread ( PTH::ThId) and is also identified by an integer identifier.

  */

  

class  SimpleRecorder : protected SimpleBufferManager, protected Binder{ 
/// count of tracer initiation
	AtomicCounter _usedCt ; // count of tracer initiation
public : 
	/// Current node id
	NodeNo _node ; // node id
/// Clock to used for timestamp of event
	 PhysicalClock * _clk; // clock used for timestamping event
/// constructor
	SimpleRecorder(NodeNo nid=0):SimpleBufferManager(), Binder(), _usedCt(0),_node(nid){};
/// destructor
	~SimpleRecorder(){}
/// open a trace recorder
SimpleBuffer *  openTrace (){    
	  SimpleBuffer * buf=allocBuffer();  
	  bindBuffer(buf);
	  return buf;
	};
/// get the current buffer with at least \b sz free size and register leve & event
	SimpleBuffer & recordEvent(tr_level_no level,  tr_event_no event,  EventSize evsz ) { 
	  SimpleBuffer * buf;
	  PhysicalClock::Date dt;
	  buf=getBuffer(); 
	  //std::cout <<" SZ ---------------> "<<evsz<<"\n";
	  if (!buf)
	    {
	      //	std::cout <<" VOID BUFF ---------------> ";
	      buf=allocBuffer(); bindBuffer(buf); 
	      //  std::cout <<" GET BUFF \n";
	    }
	  dt=_clk->getDate();
	  if(buf->_sz<evsz) 
	    { 
	      buf=flushBuffer(buf); 
	      bindBuffer(buf);
	      // register event flush buffer
	      (*buf) << tracer_level_no << flushBuffer_no << _node << buf->_tid << dt;
	      /* 
		 std::cout <<" \nLevel("  << tracer_level_no << ")::Event(" 
		 << flushBuffer_no <<  ")::Node("  
		 << _node <<   ")::Thread("  <<buf->_tid <<  ")::DATE("<<dt <<")::"; 
	      */
	      
	      dt=_clk->getDate();
	      //std::cout <<" post flush date("<< dt <<")";
	      (*buf)<<dt;
	    }
	  (*buf)<<level<<event<<_node<<buf->_tid<<dt;  
	  /*
	    std::cout <<"\n Level("  << level<< ")::Event(" <<event <<  ")::Node("  <<
	    _node <<   ")::Thread("  <<buf-> _tid<<  ")::DATE("<<dt<<")::"; 
	  */
	  return *buf;
	}
      
/// close a trace recorder. Save buffer if needed
	void closeTrace( ){ 
	  SimpleBuffer  * buf=getBuffer();
	  saveBuffer(buf);
	  buf=unBindBuffer();
	  freeBuffer(buf);
	}; 
/// Initialize a recorder with the trace file to use.
	void initialize(char * fn)throw(PosixError);
/// The recorder is on
      void commit();
/// Terminate the recorder
      void terminate();
private :


};


}
#include "tr_recorder.inl"
#endif
