// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/SOCK_net/src/SOCKNET_selector.h,v $
// Copyright(c)'2001 by Apache Team
// Authors: Thierry Gautier
//
//
// ==========================================================================
#ifndef __SOCKNET_SELECTOR_H_
#define __SOCKNET_SELECTOR_H_

#include <acom>
#include <SOCKNET_network.h>
#include <pth>
#include <sock>
#include <vector>

namespace SOCKNET {

/** High level Selector class that permits to 
    ensure thread safety of select operation
*/    
class Selector {
  Selector& operator=(const Selector& ss );
  Selector(const Selector& ss );
public:

  /** FD set kind, 'or' bits
  */
  enum SetKind {
    SET_VOID  = 0,
    SET_READ  = 1,
    SET_WRITE = 2,
    SET_EXCEP = 4,
    SET_ALL   = SET_READ | SET_WRITE | SET_EXCEP
  };

  /** Constructor
   */
  Selector();

  /** Destructor
   */
  ~Selector();

  /** initialisation
  */
  void initialize();
  
  /** destruction
  */
  void destroy();
  
  /** insert a connexion into a set 'v' of acquire-able connexions
      Modification of the set of a selector should be done
      in a exclusive way by the caller(s).
   */
  void insert(SetKind v, SOCKNET_net::Connexion* c)
     throw (ACOM::InvalidArgument);


  /** remove a connexion from a set 'v' of acquire-able connexions
      Modification of the set of a selector should be done
      in a exclusive way by the caller(s).
   */
  void remove(SetKind v, SOCKNET_net::Connexion* c)
    throw (ACOM::InvalidArgument);

  /** acquire a new active connexion.
    In case of success, returns a connexion with pending operation dependings of
    its set. The connexion is return locked for exclusive access.
    @param k (IN) the choosen set for acquiring a connexion
    @param lockset (INOUT) mutex to ensure exclusive modification of the Selector
    @param t (IN) timeout
    @retval the connexion that has been activated or 0 is return on timeout and no connexion is active
    @exception thread a ComFailure in case of error
  */
  SOCKNET_net::Connexion* acquire( SetKind v, PTH::Mutex* lockset, struct timeval* t ) 
   throw (ACOM::ComFailure);

  /** release a connexion
    @param c (IN) locked connexion to reput in the set
    @param k (IN) the set for releasing the connexion
    @param lockset (INOUT) mutex to ensure exclusive modification of the Selector
    @exception InvalidArgument if connexion is not locked, mutex invalid...
  */
  void release( SOCKNET_net::Connexion* c, SetKind v, PTH::Mutex* lockset ) 
   throw (ACOM::InvalidArgument);

  /** Force to rebuild the select sets
  */
  void force_rebuild_sets();

  /** Wakeup all thread
    @param lockset (INOUT) mutex locked by threads in critical acquire
  */
  void wakeup_all( PTH::Mutex* lockset );

protected:
  /* convert SET_READ, SET_WRITE, SET_EXCEP into 0, 1, 3 */
  int get_setindex( SetKind k ) throw (ACOM::InvalidArgument); 

  PTH::Mutex                           _access_select;
  PTH::Condition                       _condcritical;
  bool                                 _incritical;    /* true if a thread is in critical acquire */
  int                                  _waitcritical;  /* number of threads which wait to enter in critical acquire */
  bool                                 _shouldwakeup;  /* should return from acquire */
  bool                                 _force;
  int                                  _nselect;
  int                                  _fdmax;
  fd_set                               _setselected[3]; /* read, write, except FDset after select */
  fd_set                               _set[3];         /* read, write, except FDset, before select */
  std::vector<SOCKNET_net::Connexion*> _sc[3];          /* Client socket for each set */
  int                                  _pos[3];         /* next to read for each set */
  SOCK::SocketClient                   _wakeupW;        /* socket to wakeup from select */
  SOCK::SocketClient                   _wakeupR;        /* socket to wakeup from select */
};


/*
 * inline definitions
 */
inline int Selector::get_setindex( SetKind k ) throw (ACOM::InvalidArgument)
{
  switch (k) {
    case SET_READ: return 0;
    case SET_WRITE: return 1;
    case SET_EXCEP: return 2;
  }
  ACOM::throwit( ACOM::InvalidArgument());
}
 
inline Selector::Selector()
{ }
inline Selector::~Selector()
{ }

inline void Selector::insert(SetKind v, SOCKNET_net::Connexion* s)
    throw (ACOM::InvalidArgument)
{
  int sidx = get_setindex(v);
  if (_sc[sidx].size() <= (size_t)s->get_index()) 
    _sc[sidx].resize( s->get_index() +1);
  FD_SET(s->get_socket()->get_fd(),&_set[sidx]);
  _sc[sidx][s->get_index()] = s;
}

inline void Selector::remove(SetKind v, SOCKNET_net::Connexion* s)
    throw (ACOM::InvalidArgument)
{
  int sidx = get_setindex(v);
  if (_sc[sidx].size() <= (size_t)s->get_index()) 
    ACOM::throwit( ACOM::InvalidArgument() );
  FD_CLR(s->get_socket()->get_fd(),&_set[sidx]);
  _sc[sidx][s->get_index()] = 0;
}

 
} // End of namespace

#endif //end of define __SOCKNET_SELECTOR_H_
