// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/Tumit/src/conv/tr_reader.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: J. Bria t & T. Gautier

//
// ==========================================================================
#ifndef _TR_READER_H_
#define _TR_READER_H_
/*!
\file
\brief Inuktitut INRIA Recorder
\author  J. Briat 

This file is dedicated to the  reading of traces from Inuktitut INRIA tracing facility

 This part is environnement dependant and must be adapted as needed.

\date 2002
*/

#include <sys/time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>   
#include <map>
#include <list>
#include <iostream>
#include <fstream>
#include "tr.h"


using namespace std;
namespace TR 
{
  using namespace PTH;

  // exception generated when opening a  file trace failed
class FailOnOpen{};
  // exception generated when end of a file trace is got
class EndOfTrace{};

  // exception generated when an unknown event is got
class UnknownEvent{};

  // exception generated when an unknown event is got give the ev lv
class UnknownEvent2
    {
    public:
      int e,l;
      UnknownEvent2(int ev, int lv)
	{  e=ev; l=lv;}
    };

// exception generated when a bad formatted event is got
 class BadEvent{};

/** 
      \brief  A trace file class : TR::TraceFile
      
      This  class is used to overload input operator 
      to read event codes from a file.

      Reading of value is done by binary read from trace 
      file for static sized contiguous type using the 
      templated operator >>.
      This template must be specialized for specific needs. 
      This is the case for the std::string type.
*/

class TraceFile
  {
  public:
    std::fstream _f;

    TraceFile(const char * fn):_f(fn){};
    TraceFile(const char * fn, ios::openmode m):_f(fn,m){};
    //TraceFile(int fd) :_f(fd){};
    TraceFile():_f(){};
    ~TraceFile(){};
    void open(const char * fn)
      {
	_f.open( fn);
      };
    void open(const char * fn, ios::openmode m)
      {
	//_f.open( fn,m|std::ios::binary ); 
	_f.open( fn,m); 
      };
    // ios::in out ate app trunc binary
    // deffault ios::out | ios::trunc
    void close()
      {
	_f.close();
	// traitement d'erreur ?
	_f.clear();
      }
    bool is_open(){return  _f.is_open() ;}
    bool eof(){return  _f.eof();}
    bool fail(){return _f.fail();}
  };

 
// Binary input by value from TraceFile
template <class T > 
   inline TraceFile& operator>>(TraceFile& in, T* t)
   {
     in._f.read((char*)t,sizeof(T));
     return in ;
   }
 
// Binary output by value to TraceFile
template <class T > 
   inline TraceFile& operator<<(TraceFile& out, T t)
   {
     out._f.write((char*)&t,sizeof(T));
     return out ;
   }
 
// specialize the reading of a string 
inline TraceFile& operator>>(TraceFile& in, std::string *t)
   {
     PTH::StringSize sz; char * buf;
     // read the size 
     in._f.read((char*)&sz,sizeof(PTH::StringSize));
     // read in  buffer alocated on stack
     buf=(char*)alloca (sz);
     in._f.read(buf,sz);
     //then copy ...
     t->assign(buf,sz);
     return in;
   };
 
// specialize the writing of a string 
inline TraceFile& operator<<(TraceFile& out, string t)
   { 
     char * buf;
     // read the size 
     StringSize sz=t.size();
     out._f.write((char*)&sz,sizeof(PTH::StringSize));
     // read in  buffer alocated on stack
     buf=(char*)alloca (sz);
     t.copy(buf, sz);
     out._f.write(buf,sz);
     return out;
   };


/** 
     \brief  A Generic Event record  class : TR::Record 
     
     This  class  is the root class from which all record events inherit. 
     Derived record must define the date and field of event.
     
     The main use is for reading & analyzing a trace file 
 */

class Record
   {
   public:
     
     tr_level_no _lv;
     tr_event_no  _ev;
     NodeNo _node ;
     PTH::ThreadId  _tid;
     PTH::PhysicalClock::Date _d;

     Record(){};
     ~Record(){};
     Record(tr_event_no ev, tr_level_no lv): _ev(ev), _lv(lv){};
     PTH::PhysicalClock::Date getDate(){return _d;}
     int getNode(){return _node;}
     PTH::ThreadId getTid(){return _tid;};
     void getNo( tr_event_no*  ev,tr_level_no* lv){ *ev=_ev; *lv=_lv;};
     std::ostream & printHeader(std::ostream & out)
       { 
	 return out <<" Level("  << _lv << ")::Event(" <<_ev <<  ")::Node("  
		    << _node <<   ")::Thread("  << _tid<<  ")::DATE("<<_d<<")::"; 
       };
     virtual std::ostream & print(std::ostream & out)
       { return printHeader(out);}
     virtual TraceFile & read(TraceFile & in)
       { return in>>&_node>>&_tid>>&_d ;}
     virtual TraceFile & write(TraceFile & out)
       { return out<<_lv<<_ev<<_node<<_tid<<_d ;}
   };


/** 
     \brief  A Generic Event  record descriptor and registry  class : TR::GenericRecordDesc
     
     This template class implement a generic event descriptor registry  mechanism. Its goal is to register read and build procdures for each event+level code pair. After  reading in a file these codes, it is easy to build and read the corresponding event record.
     
     Template argument defines the buffer and  classes used for trace
*/
 
class RecordDesc; // forward ref
 
static const int  _lvMax=64;
static const int  _evMax=1024;
static const int  _keyMax=_lvMax * _evMax;
 
class RecordDescMap 
   {
   public:
     static int _key( tr_event_no  ev, tr_level_no  lv)
       {return ((lv%_lvMax)<<10)+(ev%_evMax);};
     
     RecordDesc * _map[_keyMax];
     RecordDescMap(){};
     ~RecordDescMap(){};
     void  dump(std::ostream& out);
     RecordDesc* get(int i){ return _map[i];};
     RecordDesc* get(tr_event_no  ev, tr_level_no  lv)
       {
	 return _map[_key(ev,lv)];
       };
     void set(int i, RecordDesc* r){  _map[i]=r;}; 
     void set(tr_event_no  ev, tr_level_no  lv, RecordDesc* r){  _map[_key(ev,lv)]=r;};
   };

class RecordDesc
   {
     static int _key( PTH::tr_event_no  ev, PTH::tr_level_no  lv)
       {return ((lv%_lvMax)<<10)+(ev%_evMax);};
   public:
     RecordDesc(){};
     RecordDesc(tr_event_no  ev, tr_level_no  lv, RecordDescMap* _rm){
       _rm->set(  _key(ev,lv),     this);
     };
     ~RecordDesc(){ };
     virtual std::ostream & print(std::ostream & out){return out;}
     virtual Record* build(TraceFile&  t )
       {/*cout<< " FAILED TO FIND THE VIRTUAL RECORD BUILDER!!!\n"; */ return 0;
       };
     /*
       static RecordDesc* getDesc(tr_event_no  ev, tr_level_no  lv){
       return _rm.get(  _key(ev,lv)   );
       };
     */
   };
 
 /** 
     \brief  A Reader class : TR::FileReader
     
     This  class implements a trace reader  mechanism.  
     Depend on global map giving event builder and event reader
 */


class  FileReader 
   {
     public : 
       RecordDescMap * _rdm;
     TraceFile    _trin ; // trace input stream
     
     
     // constructor
     FileReader(RecordDescMap * rdm):_rdm(rdm),_trin(){};
     // destructor
     ~FileReader(){}
     
     // open a trace reader for a file and a node
     void open( char *fn ) throw (FailOnOpen);
     
     // close a trace file 
     void close( ){ _trin.close();}
     
     //  read and build an event from a file
     Record* readEvent()throw ( EndOfTrace,BadEvent, UnknownEvent, UnknownEvent2); 
     
   };



 /** 
     \brief  FlushBuffer event
     
     events used to trace the behavior of the buffer allocation strategy.
 */
 
 
 class flushBuffer_rcd :public Record
   {
     public :
       flushBuffer_rcd():Record(flushBuffer_no,tracer_level_no){};
     ~flushBuffer_rcd(){};
     PTH::PhysicalClock::Date  flushBeginDate ; //  date for request to flush
     TraceFile &  read(TraceFile & in)
       { (Record::read(in))>> &flushBeginDate;};
     TraceFile &  write(TraceFile & out)
       { (Record::write(out))<<flushBeginDate;};
     std::ostream& print(std::ostream& out )
       {  
	 return  Record::print(out) <<
	   "Buffer_Management_flushBuffer(flushBeginDate=="<<flushBeginDate<<")"; 
       }
   };


 /** 
     \brief  FlushBuffer descriptor event
     
     events used to trace the behavior of the buffer allocation strategy.
 */

 
class flushBuffer_desc :public RecordDesc
   {
   public:
     flushBuffer_desc( PTH::tr_event_no ev,  PTH::tr_level_no lv, RecordDescMap * rdm  ):RecordDesc(ev,lv,rdm) { };
     ~flushBuffer_desc(){};
     //  build an event record from a file 
     flushBuffer_rcd * build(TraceFile & in);
     std::ostream & print(std::ostream & out)
       {out << "Buffer_Management_flushBuffer"; return out; };
   };
 
 // output flusBuffer record
 
 extern std::ostream& operator<<(std::ostream& out, flushBuffer_rcd& r);
 
#include "tr_reader.inl"
}
#endif
